import hashlib
import json
import logging
import requests
import re
from requests.packages.urllib3.util.retry import Retry

# Unfortunately, version of 'six' available on RHEL 6 doesn't cover this redirect
try:
    import urllib2 as request
except ImportError:  # pragma: no cover
    from urllib import request

from .exceptions import ManifestTypeError, RegistryAuthError
from .quay_session import QuaySession

LOG = logging.getLogger("pubtools.quay")


class QuayClient:
    """Class for performing Docker HTTP API operations with the Quay registry."""

    MANIFEST_LIST_TYPE = "application/vnd.docker.distribution.manifest.list.v2+json"
    MANIFEST_V2S2_TYPE = "application/vnd.docker.distribution.manifest.v2+json"

    def __init__(self, username, password, host=None):
        """
        Initialize.

        Args:
            username (str):
                Quay username.
            password (str):
                Quay password.
            host (str):
                Quay registry URL.
        """
        self.username = username
        self.password = password
        self.session = QuaySession(hostname=host, api="docker")

    def get_manifest(self, image, raw=False, manifest_list=False):
        """
        Get manifest of a given image along with its type.

        Manifest type order of preference is:
        1. manifest list
        2. V2S2 manifest
        3. anything else

        Args:
            image (str):
                Image for which to get the manifest list.
            raw (bool):
                Whether to return the manifest as raw JSON.
            manifest_list (bool):
                Whether to only return a manifest list and raise an exception otherwise.
        Returns (dict|str):
            Image manifest
                Raises:
            ManifestTypeError:
                When the image doesn't have a manifest list.
        """
        repo, ref = self._parse_and_validate_image_url(image)
        endpoint = "{0}/manifests/{1}".format(repo, ref)

        # request 'Content-Type' to be manifest list
        kwargs = {"headers": {"Accept": QuayClient.MANIFEST_LIST_TYPE}}
        response = self._request_quay("GET", endpoint, kwargs)
        if manifest_list and response.headers["Content-Type"] != QuayClient.MANIFEST_LIST_TYPE:
            raise ManifestTypeError("Image {0} doesn't have a manifest list".format(image))

        # We asked for ML but received neither ML nor V2S2. Let's ask again for V2S2
        if (
            response.headers["Content-Type"] != QuayClient.MANIFEST_LIST_TYPE
            and response.headers["Content-Type"] != QuayClient.MANIFEST_V2S2_TYPE
        ):
            kwargs = {"headers": {"Accept": QuayClient.MANIFEST_V2S2_TYPE}}
            response = self._request_quay("GET", endpoint, kwargs)

        if raw:
            return response.text
        else:
            return response.json()

    def get_manifest_digest(self, image):
        """
        Get manifest of the specified image and calculate its digest by hashing it.

        Args:
            image (str):
                Image address for which to calculate the digest.
        Returns (str):
            Manifest digest of the image.
        """
        manifest = self.get_manifest(image, raw=True)
        # SHA 256 is pretty much the standard for container images
        hasher = hashlib.sha256()
        hasher.update(manifest.encode("utf-8"))
        digest = hasher.hexdigest()

        return "sha256:{0}".format(digest)

    def upload_manifest(self, manifest, image, raw=False):
        """
        Upload manifest to a specified image.

        All manifest types are supported (manifest, manifest list).

        Args:
            manifest (dict):
                Manifest to be uploaded.
            image (str):
                Image address to upload the manifest to.
            raw (bool):
                Whether the given manifest is a string (raw) or a Python dictionary
        """
        repo, ref = self._parse_and_validate_image_url(image)
        endpoint = "{0}/manifests/{1}".format(repo, ref)

        if raw:
            manifest_type = json.loads(manifest)["mediaType"]
            kwargs = {
                "headers": {"Content-Type": manifest_type},
                "data": manifest,
            }
        else:
            manifest_type = manifest["mediaType"]
            kwargs = {
                "headers": {"Content-Type": manifest_type},
                "data": json.dumps(manifest, sort_keys=True, indent=4),
            }
        self._request_quay("PUT", endpoint, kwargs)

    def get_repository_tags(self, repository, raw=False):
        """
        Get tags of a provided repository.

        Args:
            repository (str):
                Repository whose tags should be gathered (expected format namespce/repo).
            raw (bool):
                Whether the given manifest is a string (raw) or a Python dictionary
        Returns (list):
            Tags which the repository contains.
        """
        endpoint = "{0}/tags/list".format(repository)
        response = self._request_quay("GET", endpoint)
        tags = response.json()

        while "Link" in response.headers:
            # next page response has format '</v2/....>; rel="next"'
            matches = re.findall('</v2/(.+?)>; rel="next"', response.headers["Link"])
            if len(matches) != 1:
                raise ValueError(
                    "Could not extract next page URL from response '{0}'".format(
                        response.headers["Link"]
                    )
                )
            response = self._request_quay("GET", matches[0])
            tags["tags"].extend(response.json()["tags"])

        if raw:
            return json.dumps(tags)
        else:
            return tags

    def _request_quay(self, method, endpoint, kwargs={}):
        """
        Perform a Docker HTTP API request on Quay registry. Handle authentication.

        Args:
            method (str):
                REST API method of the request (GET, POST, PUT, DELETE).
            endpoint (str):
                Endpoint of the request.
            kwargs (dict):
                Optional arguments to add to the Request object.
        Returns (Response):
            Request library's Response object.
        Raises:
            HTTPError: When the request returned an error status.
        """
        r = self.session.request(method, endpoint, **kwargs)
        # 401 is tolerated as Bearer token might need to be generated
        if r.status_code >= 400 and r.status_code < 600 and r.status_code != 401:
            r.raise_for_status()
        if r.status_code == 401:
            LOG.debug("Unauthorized request, attempting to authenticate.")
            self._authenticate_quay(r.headers)
        else:
            return r

        r = self.session.request(method, endpoint, **kwargs)
        r.raise_for_status()

        return r

    def _authenticate_quay(self, headers):
        """
        Attempt to perform an authentication with registry's authentication server.

        Once authentication is complete, add the token to the Session object.
        Specifics can be found at https://docs.docker.com/registry/spec/auth/token/

        Args:
            headers (dict):
                Headers of the 401 response received from the registry.
        Raises:
            RegistryAuthError:
                When there's an issue with the authentication procedure.
        """
        if "WWW-Authenticate" not in headers:
            raise RegistryAuthError(
                "'WWW-Authenticate' is not in the 401 response's header. "
                "Authentication cannot continue."
            )
        if "Bearer " not in headers["WWW-Authenticate"]:
            raise RegistryAuthError(
                "Different than the Bearer authentication type was requested. "
                "Only Bearer is supported."
            )

        # parse header to get a dictionary
        params = request.parse_keqv_list(
            request.parse_http_list(headers["WWW-Authenticate"][len("Bearer ") :])  # noqa: E203
        )
        host = params.pop("realm")
        session = requests.Session()
        retry = Retry(
            total=3,
            read=3,
            connect=3,
            backoff_factor=2,
            status_forcelist=set(range(500, 512)),
        )
        adapter = requests.adapters.HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)
        # Make an authentication request to the specified realm with the provided REST parameters.
        # Basic username + password authentication is expected.
        r = session.get(host, params=params, auth=(self.username, self.password))
        r.raise_for_status()

        if "token" not in r.json():
            raise RegistryAuthError("Authentication server response doesn't contain a token.")
        self.session.set_auth_token(r.json()["token"])

    def _parse_and_validate_image_url(self, image):
        """
        Extract image repository + reference from an image and validate its data.

        Args:
            image (str):
                A quay.io image.
        Returns (str, str):
            Tuple containing image repository (without base URL) and reference.
        Raises:
            ValueError:
                If image doesn't contain the expected data.
        """
        url_parts = image.split("/")
        if "@" in url_parts[-1]:
            remainder, ref = url_parts[-1].split("@")
        elif ":" in url_parts[-1]:
            remainder, ref = url_parts[-1].split(":")
        else:
            raise ValueError("Neither tag nor digest were found in the image")

        # Skip base URL and get last part of URL without reference
        repo = "/".join(url_parts[1:-1] + [remainder])
        return (repo, ref)



# this code has been copied from rcm-pub (pubd/lib/docker_signature.py)
import logging
import os
import json
import proton.handlers
import proton.reactor
import proton

import monotonic

LOG = logging.getLogger("pubtools.quay")

# There are some linting errors since this was copied from rcm-pub
# flake8: noqa: D101, D107, D102, D205, D400, D403


class MessageHandlerTimeoutException(Exception):

    pass


class ManifestClaimsHandler(proton.handlers.MessagingHandler):
    """
    Class for handling communication with RADAS via UMB.

    The manifest claims handler waits for messages from the UMB
    and will check against the manifest claims sent until all claims are received.

    - umb_urls - list of Universal Message Bus url/location
        will look like ["amqps://umb.fake.redhat.com", ...]
    - radas_address - the address/topic to listen to for RADAS messages
        e.g "topic://VirtualTopic.eng.robosignatory.container.sign"
    - claim_messages - the messages sent which will be used
        to match up responses (as you can get responses from other workers).
        This will use the request_id field
    - pub_cert - the certificate + private key combination from issued
        to connect to the Universal Message bus.
    - ca_cert - the certificate authority for the the pub_cert that
        is used to identify pub (and others) connecting to the UMB
    - timeout - the time in seconds which the message handler will
        attempt to retrieve messages. Will fail with a
        MessageHandlerTimeoutException if this is exceeded
    - message_sender_callback - this is the message_sender_callback
        that will be called with no parameters to send the messages
        to RADAS for signing. This is called after the receiver is connected
        to avoid any potential race conditions due to slow network
    """

    TIMER_TASK_DELAY = int(os.getenv("PUB_UMB_TIMER_TASK_DELAY", "10"))
    # delay between polls

    def __init__(
        self,
        umb_urls,
        radas_address,
        claim_messages,
        pub_cert,
        ca_cert,
        timeout,
        throttle,
        retry,
        message_sender_callback,
    ):
        super(ManifestClaimsHandler, self).__init__()
        self.umb_urls = umb_urls
        self.radas_address = radas_address
        self.claim_messages = list(claim_messages)
        self.received_messages = []
        self.timeout = timeout
        self.throttle = throttle
        self.retry = retry
        self.timer_task = None
        self.receiver = None
        self.message_sender_callback = message_sender_callback
        self.awaiting_response = {}  # {request_id: monotonic.monotonic()}
        self.retry_count = {}  # {request_id: 1/2}
        # a mutable list caches messages to send
        self.to_send = list(claim_messages)
        # {request_id: message} a map used to find wanted message by request_id
        self.id_msg_map = {}
        for msg in self.claim_messages:
            self.id_msg_map[msg["request_id"]] = msg
        self.connected = False

        self.ssl_domain = proton.SSLDomain(proton.SSLDomain.MODE_CLIENT)
        self.ssl_domain.set_credentials(pub_cert, pub_cert, None)
        self.ssl_domain.set_trusted_ca_db(ca_cert)
        self.ssl_domain.set_peer_authentication(proton.SSLDomain.ANONYMOUS_PEER)

    def on_start(self, event):
        LOG.debug("Message event loop starting, connecting to brokers...")
        conn = event.container.connect(
            urls=self.umb_urls, ssl_domain=self.ssl_domain, sasl_enabled=False
        )
        self.receiver = event.container.create_receiver(conn, self.radas_address)
        self.timer_task = event.container.schedule(self.timeout, self)
        # schedule a timer task, if the connection to UMB could be established, raise exception.
        LOG.debug("Message event loop started")

    def on_timer_task(self, event):
        """timer task has three functionalities:
        1. if it couldn't be connected to brokers, then after self.timeout, exception
           will be raised.
        2. Check if any request is timed out, if it is, then continue checking the retry history
            - if the retry history shows it's been tried more than 3 times, then stop loop.
            - if not, then remove it from awaiting_response dict and send it back to to_send
              queue, bump the retry count.
        3. Check if the number of requests in processing has reached throttle or not, if not
           then send corresponding number of messages.
        """
        if not self.connected:
            LOG.error("Couldn't connect to brokers after %s seconds", self.timeout)
            event.container.stop()
            raise MessageHandlerTimeoutException()
        for request_id, started in sorted(list(self.awaiting_response.items())):
            if monotonic.monotonic() - started > self.timeout:
                if request_id not in self.retry_count or self.retry_count[request_id] < self.retry:
                    self.retry_count.setdefault(request_id, 0)
                    self.retry_count[request_id] += 1
                    LOG.warn(
                        "Didn't receive response in %s for request %s, will retry [%s/%s]",
                        self.timeout,
                        request_id,
                        self.retry_count[request_id],
                        self.retry,
                    )
                    # append to resend queue and remove from awaiting_response queue
                    self.to_send.append(self.id_msg_map[request_id])
                    self.awaiting_response.pop(request_id)
                else:
                    LOG.warn("Stopping message event loop due to timeout %s", request_id)
                    event.container.stop()
                    raise MessageHandlerTimeoutException()

        # send more requests if number of waiting < throttle
        spots = self.throttle - len(self.awaiting_response)
        if self.to_send and spots > 0:
            # if there's free spots, send messages in queue
            self._send_message(min(len(self.to_send), spots))

        # schdule the next timer task
        self.timer_task = event.container.schedule(self.TIMER_TASK_DELAY, self)

    def on_link_opened(self, event):
        # do the message sending to when the connection has been opened
        # as there is the potential for messages to be dropped.
        if event.receiver == self.receiver:
            # link's open, cancel the scheduled timer task, which was for connecting
            # timing out.
            self.timer_task.cancel()
            self.connected = True
            self._send_message()
            self.timer_task = event.container.schedule(self.TIMER_TASK_DELAY, self)
        else:
            LOG.warn("Unexpected on_link_opened event")

    def on_message(self, event):
        """Once receive a message, check if it's expected by checking the awaiting_response,
        if it is, then append it to received and remove it from awaiting_response.
        """
        outer_message = json.loads(event.message.body)
        radas_message = outer_message["msg"]
        request_id = radas_message["request_id"]

        if request_id in self.awaiting_response:
            LOG.info("Received signing response: %s", request_id)
            self.awaiting_response.pop(request_id)
            self.received_messages.append(radas_message)
            if not self.to_send and not self.awaiting_response:
                LOG.info("All requests satisfied, closing connection...")
                self.receiver.close()
                event.connection.close()
                self.timer_task.cancel()
                LOG.info("Connection closed.")
        else:
            LOG.debug("Ignored signing response: %s", radas_message["request_id"])

    def on_connection_closed(self, event):
        LOG.debug("Messaging event: connection_closed")

    def on_session_closed(self, event):
        LOG.debug("Messaging event: session_closed")

    def on_link_closed(self, event):
        LOG.debug("Messaging event: link_closed")

    def on_connection_closing(self, event):
        LOG.debug("Messaging event: connection_closing")

    def on_session_closing(self, event):
        LOG.debug("Messaging event: session_closing")

    def on_link_closing(self, event):
        LOG.debug("Messaging event: link_closing")

    def on_disconnected(self, event):
        LOG.debug("Messaging event: disconnected")

    def _send_message(self, count=None):
        if count is None:
            count = min(len(self.to_send), self.throttle)
        messages = self.to_send[:count]
        LOG.info("Sending %s messages...", count)
        self.message_sender_callback(messages)
        # remove sent message from sending queue
        del self.to_send[:count]

        # add sent message's request id to waiting queue
        for msg in messages:
            self.awaiting_response[msg["request_id"]] = monotonic.monotonic()



import logging


class Logger(object):
    """Logging helper class."""

    def __init__(self):
        """Init the logger instance."""
        self.logger = logging.getLogger("pubtools.quay")

    def log_info(self, *args, **kwargs):
        """Log message with info level."""
        self.logger.log(logging.INFO, *args, **kwargs)

    def log_error(self, *args, **kwargs):
        """Log message with error level."""
        self.logger.log(logging.ERROR, *args, **kwargs)

    def log_warning(self, *args, **kwargs):
        """Log message with warning level."""
        self.logger.log(logging.WARNING, *args, **kwargs)

    def log_debug(self, *args, **kwargs):
        """Log message with debug level."""
        self.logger.log(logging.DEBUG, *args, **kwargs)


def task_status(event):
    """Log task status.

    Expand as necessary.
    """
    return dict(event={"type": event})


def log_jsonl(step_name):
    """Log task status before and after decorated method.

    For methods which constitute an entire task step.

    :param step_name: Name of the task step, e.g., "Tag images".
    """
    event_name = step_name.lower().replace(" ", "-")
    logger = logging.getLogger("pubtools.quay")

    def decorate(fn):
        def fn_wrapper(*args, **kwargs):
            try:
                logger.info("%s: Started", step_name, extra=task_status("%s-start" % event_name))
                ret = fn(*args, **kwargs)
                logger.info("%s: Finished", step_name, extra=task_status("%s-end" % event_name))
                return ret
            except Exception:
                logger.error("%s: Failed", step_name, extra=task_status("%s-error" % event_name))
                raise

        return fn_wrapper

    return decorate



import logging
import re
import yaml

import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from .container_image_pusher import ContainerImagePusher
from .utils.misc import (
    run_entrypoint,
    get_internal_container_repo_name,
    log_step,
    get_pyxis_ssl_paths,
)

LOG = logging.getLogger("pubtools.quay")


class OperatorPusher:
    """
    Add operator bundles to index images and push them to Quay.

    No validation is performed, push items are expected to be correct.
    """

    def __init__(self, push_items, target_settings):
        """
        Initialize.

        Args:
            push_items ([ContainerPushItem]):
                List of push items.
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.target_settings = target_settings

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")
        self._version_items_mapping = {}

    @staticmethod
    def _get_immutable_tag(push_item):
        """
        Return immutable tag from operator push item production tags.

        Args:
            push_item (ContainerPushItem):
                Operator push item.

        Returns (str):
            Immutable tag
        """
        # NOTE: while this was copied from pub, '[0]' had to be added here to work properly
        if push_item.metadata["v_r"] in list(push_item.metadata["tags"].values())[0]:
            return push_item.metadata["v_r"]
        # if v_r tag is not in destination tags, return tag with the most numbers in it
        # This code is usually trigerred when push is not initiated from ET but manually
        # by user
        tags = []
        for tag in list(push_item.metadata["tags"].values())[0]:
            tags.append((len(re.split(r"\d+", tag)), tag))

        return sorted(tags)[-1][1]

    def public_bundle_ref(self, push_item):
        """
        Get public reference of a bundle image.

        It will be used by IIB to access the bundle image.

        Args:
            push_item (ContainerPushItem):
                Operator push item.

        Returns (str):
            Customer-visible bundle reference.
        """
        repository = list(push_item.metadata["tags"].keys())[0]
        # tags are the same for each destination repo, so any combination should work
        return "{0}/{1}:{2}".format(
            self.target_settings["docker_settings"]["docker_reference_registry"][0],
            repository,
            self._get_immutable_tag(push_item),
        )

    def pyxis_get_ocp_versions(self, push_item):
        """
        Get a list of supported ocp versions from Pyxis.

        Args:
            push_item: (ContainerPushItem)
                Push item for which the OCP version range will be found out.

        Returns ([str]):
            Supported OCP versions as returned by Pyxis.
        """
        cert, key = get_pyxis_ssl_paths(self.target_settings)

        ocp_versions = push_item.metadata["com.redhat.openshift.versions"]
        LOG.info("Getting OCP versions of '{0}' from Pyxis.".format(ocp_versions))

        args = ["--pyxis-server", self.target_settings["pyxis_server"]]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]
        args += ["--organization", self.target_settings["iib_organization"]]
        args += ["--ocp-versions-range", ocp_versions]
        env_vars = {}

        data = run_entrypoint(
            ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-operator-indices"),
            "pubtools-pyxis-get-operator-indices",
            args,
            env_vars,
        )

        if not data:
            msg = "Pyxis has returned no OCP versions for '{0}' specified in build {1}.".format(
                ocp_versions, push_item.metadata["build"]["build_id"]
            )
            raise ValueError(msg)

        # Versions returned by Pyxis don't contain 'v' at the front (4.5 -> v4.5)
        return ["v{0}".format(item["ocp_version"]) for item in data]

    @property
    def version_items_mapping(self):
        """
        Generate mapping of OCP version -> push_items.

        The mapping describes which operator bundles should be added to which index images.

        Returns ({str: [ContainerPushItem]})
            Mapping of OCP version -> Push items
        """
        if not self._version_items_mapping:
            ocp_versions_resolved = {}

            for item in self.push_items:
                ocp_versions = item.metadata["com.redhat.openshift.versions"]
                # we haven't yet encountered this pattern, contact Pyxis for resolution
                if ocp_versions not in ocp_versions_resolved:
                    ocp_versions_resolved[ocp_versions] = self.pyxis_get_ocp_versions(item)

                for version in ocp_versions_resolved[ocp_versions]:
                    self._version_items_mapping.setdefault(version, []).append(item)

        return self._version_items_mapping

    def get_deprecation_list(self, version):
        """
        Get bundles to be deprecated in the index image.

        Args:
            version: (str)
                version for which deprecation list will be fetched.

        Returns:
            list(str): list of bundles to be deprecated in the index image.
        """

        def _get_requests_session():
            session = requests.Session()
            retry = Retry(
                total=6,
                read=6,
                connect=6,
                backoff_factor=0.8,
                status_forcelist=(500, 502, 503, 504),
            )
            adapter = HTTPAdapter(max_retries=retry)
            session.mount("http://", adapter)
            session.mount("https://", adapter)

            return session

        deprecation_list = []
        deprecation_list_url = "{0}/{1}.yml/raw?ref=master".format(
            self.target_settings["iib_deprecation_list_url"].rstrip("/"), version.replace(".", "_")
        )
        registry_url = self.target_settings["docker_settings"]["docker_reference_registry"][0]

        LOG.info("Getting the deprecation list for OCP version {0}".format(version))
        session = _get_requests_session()
        response = session.get(url=deprecation_list_url)
        if not response.ok:
            LOG.error(
                "Could not retrieve deprecation list after multiple attempts."
                " Status Code {0}".format(response.status_code)
            )
            response.raise_for_status()

        try:
            yaml_response = yaml.safe_load(response.text)
            if yaml_response:
                deprecation_list = [
                    "{0}/{1}".format(registry_url, bundle_path)
                    for pkg_deprecation_list in yaml_response.values()
                    for bundle_path in pkg_deprecation_list
                ]
        except Exception:
            LOG.error("Data in {0} is invalid".format(deprecation_list_url))
            raise

        LOG.info("Deprecation list retrieved successfully")
        return sorted(deprecation_list)

    @classmethod
    def pubtools_iib_get_common_args(cls, target_settings):
        """
        Create an argument list common for all pubtools-iib operations.

        Target settings are used to set the values of the arguments

        Args:
            target_settings (dict):
                Settings used for setting the value of pubtools-iib parameters.
        Returns (([str]), {str:str}):
            Tuple of arguments and environment variables to be used when calling pubtools-iib.
        """
        args = ["--skip-pulp"]

        args += ["--iib-server", target_settings["iib_server"]]
        args += ["--iib-krb-principal", target_settings["iib_krb_principal"]]

        if "iib_overwrite_from_index" in target_settings:
            args += ["--overwrite-from-index"]
        if "iib_krb_ktfile" in target_settings:
            args += ["--iib-krb-ktfile", target_settings["iib_krb_ktfile"]]

        env_vars = {}
        if "iib_overwrite_from_index_token" in target_settings:
            env_vars["OVERWRITE_FROM_INDEX_TOKEN"] = target_settings[
                "iib_overwrite_from_index_token"
            ]

        return (args, env_vars)

    @classmethod
    def iib_add_bundles(
        cls, bundles=None, archs=None, index_image=None, deprecation_list=None, target_settings={}
    ):
        """
        Construct and execute pubtools-iib command to add bundles to index image.

        Args:
            bundles ([str]):
                External URLs to bundle images to be added to the index image.
            archs ([str]):
                Architectures to build for.
            index_image (str):
                Index image to add the bundles to.
            deprecation_list ([str]|str):
                List of bundles to be deprecated. Accepts both str (csv) and a list.
            target_settings (dict):
                Settings used for setting the value of pubtools-iib parameters.

        Returns (dict):
            Build details provided by IIB.
        """
        LOG.info(
            "Requesting IIB to add bundles '{0}' to index image '{1}'".format(bundles, index_image)
        )
        args, env_vars = cls.pubtools_iib_get_common_args(target_settings)

        if index_image:
            args += ["--index-image", index_image]
        if bundles:
            for bundle in bundles:
                args += ["--bundle", bundle]
        if archs:
            for arch in archs:
                args += ["--arch", arch]
        # inconsistent way of presenting multiple arguments...
        if deprecation_list and isinstance(deprecation_list, str):
            args += ["--deprecation-list", deprecation_list]
        elif deprecation_list and isinstance(deprecation_list, list):
            args += ["--deprecation-list", ",".join(deprecation_list)]

        return run_entrypoint(
            ("pubtools-iib", "console_scripts", "pubtools-iib-add-bundles"),
            "pubtools-iib-add-bundles",
            args,
            env_vars,
        )

    @classmethod
    def iib_remove_operators(cls, operators=None, archs=None, index_image=None, target_settings={}):
        """
        Construct and execute pubtools-iib command to remove operators from index image.

        Args:
            operators ([str]):
                Operator names to be removed from the index image.
            archs ([str]):
                Architectures to build for.
            ocp_version (str):
                Index image to remove the operators from.
            target_settings (dict):
                Settings used for setting the value of pubtools-iib parameters.

        Returns (dict):
            Build details provided by IIB.
        """
        LOG.info(
            "Requesting IIB to remove operators '{0}' from index image '{1}'".format(
                operators, index_image
            )
        )
        args, env_vars = cls.pubtools_iib_get_common_args(target_settings)

        if index_image:
            args += ["--index-image", index_image]
        if operators:
            for operator in operators:
                args += ["--operator", operator]
        if archs:
            for arch in archs:
                args += ["--arch", arch]

        return run_entrypoint(
            ("pubtools-iib", "console_scripts", "pubtools-iib-remove-operators"),
            "pubtools-iib-remove-operators",
            args,
            env_vars,
        )

    def get_existing_index_images(self, quay_client):
        """
        Return existing index images for push items.

        Args:
            quay_client (QuayClient): quay_client_instance

        Returns [(digest, tag)]:
            List of tuples containing digest and tag of existing index image
        """
        image_schema = "{host}/{namespace}/{repo}"
        iib_repo = get_internal_container_repo_name(
            self.target_settings["quay_operator_repository"]
        )
        index_image_repo = image_schema.format(
            host=self.quay_host, namespace=self.target_settings["quay_namespace"], repo=iib_repo
        )
        current_index_images = []

        manifest_list = {}
        for version in sorted(self.version_items_mapping.keys()):
            image_ref = "{0}:{1}".format(index_image_repo, version)
            try:
                manifest_list = quay_client.get_manifest(image_ref, manifest_list=True)
            except requests.exceptions.HTTPError as e:
                if e.response.status_code == 404 or e.response.status_code == 401:
                    continue
                else:
                    raise
            for manifest in manifest_list["manifests"]:
                current_index_images.append(
                    (manifest["digest"], version, self.target_settings["quay_operator_repository"])
                )
        return list(set(current_index_images))

    @log_step("Build index images")
    def build_index_images(self):
        """
        Perform the 'build' part of the operator workflow.

        This workflow is a part of push-docker operation.
        The workflow can be summarized as:
        - Use Pyxis to parse 'com.redhat.openshift.versions'
        - Get deprecation list for a given version (list of bundles to be deprecated)
        - Create mapping of which bundles should be pushed to which index image versions
        - Contact IIB to add the bundles to the index images

        Returns ({str:dict}):
            Dictionary containing IIB results and signing keys for all OPM versions. Data will be
            used in operator signing. Dictionary structure:
            {
                "version": {
                    "iib_result": (...) (object returned by iiblib)
                    "signing_keys": [...] (list of signing keys to be used for signing)
                }
            }

        """
        iib_results = {}

        for version, items in sorted(self.version_items_mapping.items()):
            bundles = [self.public_bundle_ref(i) for i in items]
            all_archs = [
                i.metadata["arch"] if i.metadata["arch"] != "x86_64" else "amd64" for i in items
            ]
            archs = sorted(list(set(all_archs)))
            signing_keys = sorted(list(set([item.claims_signing_key for item in items])))

            # Get deprecation list
            deprecation_list = self.get_deprecation_list(version)

            # build index image in IIB
            index_image = "{image_repo}:{tag}".format(
                image_repo=self.target_settings["iib_index_image"], tag=version
            )
            build_details = self.iib_add_bundles(
                bundles=bundles,
                archs=archs,
                index_image=index_image,
                deprecation_list=deprecation_list,
                target_settings=self.target_settings,
            )

            iib_results[version] = {"iib_result": build_details, "signing_keys": signing_keys}

        return iib_results

    @log_step("Push index images to Quay")
    def push_index_images(self, iib_results):
        """
        Push index images which were built in the previous stage to Quay.

        Args:
            iib_results (dict):
                IIB results returned by the build stage
        """
        image_schema = "{host}/{namespace}/{repo}"
        index_image_repo = image_schema.format(
            host=self.quay_host,
            namespace=self.target_settings["quay_namespace"],
            repo=get_internal_container_repo_name(self.target_settings["quay_operator_repository"]),
        )

        for version in self.version_items_mapping:
            build_details = iib_results[version]["iib_result"]

            _, tag = build_details.index_image.split(":", 1)
            dest_image = "{0}:{1}".format(index_image_repo, tag)
            ContainerImagePusher.run_tag_images(
                build_details.index_image, [dest_image], True, self.target_settings
            )



import logging

from pubtools.pluggy import task_context

from .utils.misc import setup_arg_parser, add_args_env_variables
from .manifest_list_merger import ManifestListMerger

LOG = logging.getLogger("pubtools.quay")

MERGE_MANIFEST_LIST_ARGS = {
    ("--source-ref",): {
        "help": "Source image reference. Manifest list data of this image will "
        "overwrite destination's manifest list",
        "required": True,
        "type": str,
    },
    ("--dest-ref",): {
        "help": "Destination image reference. Must be specified by tag. New manifest list will be "
        "uploaded to this image reference.",
        "required": True,
        "type": str,
    },
    ("--source-quay-user",): {
        "help": "Quay username to get source image.",
        "required": True,
        "type": str,
    },
    ("--source-quay-password",): {
        "help": "Quay password to get source image. "
        "Can be specified by env variable SOURCE_QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "SOURCE_QUAY_PASSWORD",
    },
    ("--dest-quay-user",): {
        "help": "Quay username to get dest image.",
        "required": True,
        "type": str,
    },
    ("--dest-quay-password",): {
        "help": "Quay password to get dest image. "
        "Can be specified by env variable DEST_QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "DEST_QUAY_PASSWORD",
    },
}


def verify_merge_manifest_list_args(args):
    """Verify the presence and correctness of input parameters."""
    if "@" in args.dest_ref:
        raise ValueError("Destination must be specified via tag, not digest")

    if not args.source_quay_password or not args.dest_quay_password:
        raise ValueError("Quay password must be set for both source and dest images")


def setup_args():
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(MERGE_MANIFEST_LIST_ARGS)


def merge_manifest_list_main(sysargs=None):
    """Entrypoint for manifest list merging."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, MERGE_MANIFEST_LIST_ARGS)

    verify_merge_manifest_list_args(args)
    merger = ManifestListMerger(
        args.source_ref,
        args.dest_ref,
        args.source_quay_user,
        args.source_quay_password,
        args.dest_quay_user,
        args.dest_quay_password,
    )

    with task_context():
        merger.merge_manifest_lists()



# Compatibility helpers for caplog in py2 vs py3.
#
# Note: delete this file when python2 is dropped!


class CapturelogWrapper(object):
    """A wrapper for pytest-catchlog fixture (for python2.6).

    This allows many tests to use caplog's modern API without having
    to add branches to care about py2 support.

    Note that this class is not guaranteed to make everything behave
    identically though. You should update it if your test needs
    something missing here.
    """

    def __init__(self, legacy):
        """
        Initilize.

        Args:
            legacy:
                Legacy caplog fixture.
        """
        self._legacy = legacy

        # These are identical methods between old and new APIs
        self.set_level = legacy.set_level
        self.at_level = legacy.at_level

    @property
    def messages(self):
        """
        Property in caplog, does not exist at all in legacy fixture.

        This property gives the interpolated log messages only, without
        applying the logger's configured formatter (e.g. "something happened"
        and not "[ERROR] something happened").
        """
        return [rec.getMessage() for rec in self.records]

    # These are identical properties between old and new APIs
    handler = property(lambda self: self._legacy.handler)
    text = property(lambda self: self._legacy.text)
    records = property(lambda self: self._legacy.records)



import logging

from .container_image_pusher import ContainerImagePusher
from .exceptions import InvalidTargetSettings
from .operator_pusher import OperatorPusher
from .signature_handler import OperatorSignatureHandler
from .signature_remover import SignatureRemover
from .utils.misc import get_internal_container_repo_name, get_pyxis_ssl_paths

LOG = logging.getLogger("pubtools.quay")


def verify_target_settings(target_settings):
    """
    Verify the presence and validity of target settings.

    Args:
        target_settings (dict):
            Dictionary containing settings necessary for performing the operation.
    """
    LOG.info("Verifying the necessary target settings")

    required_settings = [
        "source_quay_user",
        "source_quay_password",
        "dest_quay_user",
        "dest_quay_password",
        "dest_quay_api_token",
        "pyxis_server",
        "quay_namespace",
        "iib_krb_principal",
        "iib_index_image",
        "quay_operator_repository",
        "skopeo_image",
        "docker_settings",
    ]

    for setting in required_settings:
        if setting not in target_settings:
            raise InvalidTargetSettings(
                "'{0}' must be present in the target settings.".format(setting)
            )

    required_docker_settings = ["umb_urls", "docker_reference_registry"]
    for setting in required_docker_settings:
        if setting not in target_settings["docker_settings"]:
            raise InvalidTargetSettings(
                "'{0}' must be present in the docker settings.".format(setting)
            )
    if (
        "iib_overwrite_from_index_token" in target_settings
        and "iib_overwrite_from_index" not in target_settings
    ) or (
        "iib_overwrite_from_index_token" not in target_settings
        and "iib_overwrite_from_index" in target_settings
    ):
        msg = (
            "Either both or neither of 'iib_overwrite_from_index' and "
            "'iib_overwrite_from_index_token' should be specified in target settings."
        )
        LOG.error(msg)
        raise InvalidTargetSettings(msg)


def task_iib_add_bundles(
    bundles,
    archs,
    index_image,
    deprecation_list,
    signing_keys,
    hub,
    task_id,
    target_settings,
    target_name,
):
    """
    Perform all the necessary actions for the 'PushAddIIBBundles' entrypoint.

    Args:
        bundles ([str]):
            Bundles to add to the index image.
        archs ([str]):
            Architectures to build the index image for.
        index_image (str):
            Index image to add the bundles to.
        deprecation_list ([str]):
            Bundles to deprecate in the index image.
        signing_keys ([str]):
            Signing keys to be used.
        hub (HubProxy):
            Instance of XMLRPC pub-hub proxy.
        task_id (str):
            ID of the pub task.
        target_settings (dict):
            Dictionary containing settings necessary for performing the operation.
        target_name (str):
            Name of the target.
    """
    image_schema_tag = "{host}/{namespace}/{repo}:{tag}"
    image_schema_digest = "{host}/{namespace}/{repo}@{digest}"
    verify_target_settings(target_settings)

    # Build new index image in IIB
    build_details = OperatorPusher.iib_add_bundles(
        bundles=bundles,
        archs=archs,
        index_image=index_image,
        deprecation_list=deprecation_list,
        target_settings=target_settings,
    )

    _, tag = build_details.index_image.split(":", 1)
    dest_image = image_schema_tag.format(
        host=target_settings.get("quay_host", "quay.io").rstrip("/"),
        namespace=target_settings["quay_namespace"],
        repo=get_internal_container_repo_name(target_settings["quay_operator_repository"]),
        tag=tag,
    )
    # Index image used to fetch manifest list. This image will never be overwritten
    iib_namespace = build_details.index_image_resolved.split("/")[1]
    image_digest = build_details.index_image_resolved.split("@")[1]
    intermediate_index_image = image_schema_digest.format(
        host=target_settings.get("quay_host", "quay.io").rstrip("/"),
        namespace=iib_namespace,
        repo="iib",
        digest=image_digest,
    )

    # Sign image
    sig_handler = OperatorSignatureHandler(hub, task_id, target_settings, target_name)
    claim_messages = sig_handler.sign_task_index_image(signing_keys, intermediate_index_image, tag)

    sig_remover = SignatureRemover(
        quay_api_token=target_settings["dest_quay_api_token"],
        quay_user=target_settings["dest_quay_user"],
        quay_password=target_settings["dest_quay_password"],
    )
    cert, key = get_pyxis_ssl_paths(target_settings)

    old_signatures = sig_remover.get_index_image_signatures(
        dest_image,
        claim_messages,
        target_settings["pyxis_server"],
        cert,
        key,
    )

    # Push image to Quay
    # NOTE: tagging doesn't use intermediate index image, because we want the most up-to-date
    #       image to be copied to the destination
    ContainerImagePusher.run_tag_images(
        build_details.index_image, [dest_image], True, target_settings
    )

    signature_ids = [s["_id"] for s in old_signatures]
    sig_remover.remove_signatures_from_pyxis(
        signature_ids,
        target_settings["pyxis_server"],
        cert,
        key,
    )


def task_iib_remove_operators(
    operators, archs, index_image, signing_keys, hub, task_id, target_settings, target_name
):
    """
    Perform all the necessary actions for the 'PushRemoveIIBOperators' entrypoint.

    Args:
        operators ([str]):
            Operators to remove from the index image.
        arch ([str]):
            Architectures to build the index image for.
        index_image (str):
            Index image to remove the operators from.
        signing_keys (str):
            Signing keys to be used.
        hub (HubProxy):
            Instance of XMLRPC pub-hub proxy.
        task_id (str):
            ID of the pub task.
        target_settings (dict):
            Dictionary containing settings necessary for performing the operation.
        target_name (str):
            Name of the target.
    """
    image_schema_tag = "{host}/{namespace}/{repo}:{tag}"
    image_schema_digest = "{host}/{namespace}/{repo}@{digest}"
    verify_target_settings(target_settings)

    # Build new index image in IIB
    build_details = OperatorPusher.iib_remove_operators(
        operators=operators,
        archs=archs,
        index_image=index_image,
        target_settings=target_settings,
    )

    _, tag = build_details.index_image.split(":", 1)
    dest_image = image_schema_tag.format(
        host=target_settings.get("quay_host", "quay.io").rstrip("/"),
        namespace=target_settings["quay_namespace"],
        repo=get_internal_container_repo_name(target_settings["quay_operator_repository"]),
        tag=tag,
    )

    # Index image used to fetch manifest list. This image will never be overwritten
    iib_namespace = build_details.index_image_resolved.split("/")[1]
    image_digest = build_details.index_image_resolved.split("@")[1]
    intermediate_index_image = image_schema_digest.format(
        host=target_settings.get("quay_host", "quay.io").rstrip("/"),
        namespace=iib_namespace,
        repo="iib",
        digest=image_digest,
    )

    # Sign image
    sig_handler = OperatorSignatureHandler(hub, task_id, target_settings, target_name)
    claim_messages = sig_handler.sign_task_index_image(signing_keys, intermediate_index_image, tag)

    sig_remover = SignatureRemover(
        quay_api_token=target_settings["dest_quay_api_token"],
        quay_user=target_settings["dest_quay_user"],
        quay_password=target_settings["dest_quay_password"],
    )
    cert, key = get_pyxis_ssl_paths(target_settings)

    old_signatures = sig_remover.get_index_image_signatures(
        dest_image,
        claim_messages,
        target_settings["pyxis_server"],
        cert,
        key,
    )

    # Push image to Quay
    ContainerImagePusher.run_tag_images(
        build_details.index_image, [dest_image], True, target_settings
    )

    signature_ids = [s["_id"] for s in old_signatures]
    sig_remover.remove_signatures_from_pyxis(
        signature_ids,
        target_settings["pyxis_server"],
        cert,
        key,
    )


def task_iib_build_from_scratch(
    bundles, archs, index_image_tag, signing_keys, hub, task_id, target_settings, target_name
):
    """
    Perform all the necessary actions for the 'PushIIBBuildFromScratch' entrypoint.

    Args:
        bundles ([str]):
            Bundles to add to the index image.
        archs ([str]):
            Architectures to build the index image for.
        index_image_tag (str):
            Tag to be applied to the new index image.
        signing_keys (str):
            Signing keys to be used.
        hub (HubProxy):
            Instance of XMLRPC pub-hub proxy.
        task_id (str):
            ID of the pub task.
        target_settings (dict):
            Dictionary containing settings necessary for performing the operation.
        target_name (str):
            Name of the target.
    """
    image_schema_tag = "{host}/{namespace}/{repo}:{tag}"
    image_schema_digest = "{host}/{namespace}/{repo}@{digest}"
    verify_target_settings(target_settings)

    # Build new index image in IIB
    build_details = OperatorPusher.iib_add_bundles(
        bundles=bundles,
        archs=archs,
        target_settings=target_settings,
    )

    dest_image = image_schema_tag.format(
        host=target_settings.get("quay_host", "quay.io").rstrip("/"),
        namespace=target_settings["quay_namespace"],
        repo=get_internal_container_repo_name(target_settings["quay_operator_repository"]),
        tag=index_image_tag,
    )

    _, tag = build_details.index_image.split(":", 1)
    # Index image used to fetch manifest list. This image will never be overwritten
    iib_namespace = build_details.index_image_resolved.split("/")[1]
    image_digest = build_details.index_image_resolved.split("@")[1]
    intermediate_index_image = image_schema_digest.format(
        host=target_settings.get("quay_host", "quay.io").rstrip("/"),
        namespace=iib_namespace,
        repo="iib",
        digest=image_digest,
    )

    # Sign image
    sig_handler = OperatorSignatureHandler(hub, task_id, target_settings, target_name)
    sig_handler.sign_task_index_image(signing_keys, intermediate_index_image, index_image_tag)

    # Push image to Quay
    ContainerImagePusher.run_tag_images(
        build_details.index_image, [dest_image], True, target_settings
    )


def iib_add_entrypoint(
    bundles,
    archs,
    index_image,
    deprecation_list,
    signing_keys,
    hub,
    task_id,
    target_settings,
    target_name,
):
    """Entry point for use in another python code."""
    task_iib_add_bundles(
        bundles,
        archs,
        index_image,
        deprecation_list,
        signing_keys,
        hub,
        task_id,
        target_settings,
        target_name,
    )


def iib_remove_entrypoint(
    operators, archs, index_image, signing_keys, hub, task_id, target_settings, target_name
):
    """Entry point for use in another python code."""
    task_iib_remove_operators(
        operators, archs, index_image, signing_keys, hub, task_id, target_settings, target_name
    )


def iib_from_scratch_entrypoint(
    bundles, archs, index_image_tag, signing_keys, hub, task_id, target_settings, target_name
):
    """Entry point for use in another python code."""
    task_iib_build_from_scratch(
        bundles, archs, index_image_tag, signing_keys, hub, task_id, target_settings, target_name
    )



from __future__ import print_function
import re
import six
import pkg_resources

from .utils.stepper import Step, StepFailedError
from .utils.logger import log_jsonl


LOG_INDENT = "  - "


class StepSanitizeContainerPushItems(Step):
    """Filter push items with error state or with empty file_path and ignore non docker items.

    Expected step args:
        ()
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
        "push_items": list of push items
    Produced details:
        {"item": str "state": (ready, not-container, error, no-pull-data, ok)}
    Results:
        list of indexes of container push items in push_items list
    """

    NAME = "StepSanitizeContainerPushItems"

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        self._details = []
        for item in push_items:
            self._details.append({"item": str(item), "state": "ready"})

    def _update_details(self, details):
        index, state = details
        self._details[index]["state"] = state

    @log_jsonl("SanitizeContainerPushItems")
    def _run(self, on_update=None):
        push_items = self.external_resources["push_items"]
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        log_warning = self.external_resources.get("log_warning", lambda *args, **kwargs: ())
        log_error = self.external_resources.get("log_error", lambda *args, **kwargs: ())

        container_push_items = []
        _on_update = on_update if on_update else lambda: ()
        failed = False
        for i, item in enumerate(push_items):
            log_info("Processing %s", str(item))
            item_error = False
            item_state = "ok"
            if item.file_type != "docker":
                log_warning(LOG_INDENT + "Item %s is not container, skipping" % str(item))
                self.update_details((i, "not-container"))
                _on_update()
                continue
            if item.file_path is None:
                item_state = "error"
                log_error(LOG_INDENT + "File path of %s is None" % str(item))
                self.results.errors.setdefault("item_errors", []).append(
                    (str(item), "empty file_path")
                )
                item_error = True
                failed = True
                if item.state is None:
                    item.state = "NOTPUSHED"
            if item.errors:
                item_state = "error"
                self.results.errors.setdefault("item_errors", []).append((str(item), item.errors))
                for message in item.errors.values():
                    log_error(LOG_INDENT + "Bad push item '%s': %s" % (item.file_name, message))
                item.state = "INVALIDFILE"
                item_error = True
                failed = True
            if not item.metadata.get("pull_data"):
                item_state = "no-pull-data"
                self.results.errors.setdefault("item_errors", []).append(
                    (str(item), "Cannot calculate pull data")
                )
                log_error(LOG_INDENT + "Missing pull_data for %s" % str(item))
                item_error = True
                failed = True
            if not item_error:
                log_info(LOG_INDENT + "Processing %s ok" % str(item))
            self.update_details((i, item_state))
            _on_update()

            container_push_items.append(i)
        if failed:
            raise StepFailedError()
        self.results.results = container_push_items


class StepSanitizeOperatorPushItems(Step):
    """Filter out push items which are not operators and store result.

    Expected step args:
        ()
    Expected step kw args:
        {"auto_upload_operators": <True|False>}
    Produced details:
        item -> {"item": str, "state": (ready, not-operator, unsupported-legacy,
                                        unknown-op-type, no-ocp-version, ok)}
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
        "push_items": list of push items
    Results:
        list of indexes of container push items in push_items list
    """

    NAME = "StepSanitizeOperatorPushItems"

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        self._details = []
        for item in push_items:
            self._details.append({"item": str(item), "state": "ready"})

    def _update_details(self, details):
        index, state = details
        self._details[index]["state"] = state

    def _pre_run(self):
        if not self.step_kwargs.get("auto_upload_operators", True):
            self.skip = True
            self.skip_reason = "Automatic uploading of operators is not enabled"

    @log_jsonl("SanitizeOperatorPushItems")
    def _run(self, on_update=None):
        push_items = self.external_resources["push_items"]
        log_debug = self.external_resources.get("log_debug", lambda *args, **kwargs: ())
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        log_error = self.external_resources.get("log_error", lambda *args, **kwargs: ())
        operator_push_items = []
        _on_update = on_update if on_update else lambda: ()
        errors = False
        for i, item in enumerate(push_items):
            log_debug("Processing %s", str(item))
            _on_update()
            if item.file_type != "operator":
                log_info(LOG_INDENT + "Item %s is not operator, skipping", str(item))
                item_state = "not-operator"
                self.update_details((i, item_state))
                on_update()
                continue

            item_state = "ok"
            op_type = item.metadata.get("op_type") or "appregistry"
            ocp_versions = item.metadata.get("com.redhat.openshift.versions")
            if op_type == "appregistry":
                item_state = "unsupported-legacy"
                log_info(
                    LOG_INDENT + "Item %s is unsupported legacy (appregistry), skipping",
                    str(item),
                )
            elif op_type != "bundle":
                item_state = "unknown-op-type"
                self.results.errors.setdefault("item_errors", []).append(
                    (str(item), "unknown operator type: %s" % op_type)
                )
                log_error(
                    LOG_INDENT + "Item %s has unknown operator type: %s",
                    op_type,
                    str(item),
                )
                errors = True
            elif not ocp_versions:
                item_state = "no-ocp-version"
                self.results.errors.setdefault("item_errors", []).append(
                    (
                        str(item),
                        "'com.redhat.openshift.versions' is not specified for build",
                    )
                )
                log_error(
                    LOG_INDENT + "Item %s is missing 'com.redhat.openshift.versions' label",
                    str(item),
                )
                errors = True
            if item_state == "ok":
                operator_push_items.append(i)
                log_debug(LOG_INDENT + "Processing %s ok", str(item))
            self.update_details((i, item_state))
            _on_update()
        if errors:
            raise StepFailedError()
        self.results.results = operator_push_items


class StepSanitizeRepositories(Step):
    """Sanitize destination repositories.

    Expected step args:
        (<key-for-container-items-indexes>,)
    Produced details:
        repository -> (ready, not-exists, depracated, ok)
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
        "push_items": list of push items
    """

    NAME = "StepSanitizeRepositories"

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        self._details = {}
        repositories = set()
        container_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[container_items_indexes_key].results
        for index in indexes:
            item = push_items[index]
            for repo in item.repos:
                repositories.add(repo)
        for repo in repositories:
            self._details[repo] = "ready"

    def _sanitize_repository(self, repo):
        raise NotImplementedError

    @log_jsonl("SanitizeOperatorPushItems")
    def _run(self, on_update=None):
        for repo in self._details.keys():
            self._sanitize_repository(repo)


class StepBuildBackupMapping(Step):
    """Build repo -> [backup tags] mapping repo -> [rollback tags].

    Expected step args:
        (<key-for-container-items-indexes>,)
    Produced details:
        {
            "backup_tags": { <repo>: { <tag>: <backup> }}
            "rollback_tags": { <repo>: {<tag>: <state>}}
        }
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
        "push_items": list of push items
    Results:
        {
            "backup_tags": { <repo>: { <tag>: <backup> }}
            "rollback_tags": { <repo>: {<tag>: <state>}}
        }
    """

    NAME = "StepBuildBackupMapping"

    def _update_tag_backups(self, backup_tags):
        # TODO:
        # 1. get existing tags from quay and put them as backup
        # 2. put non-existing tags to rollback tags
        raise NotImplementedError

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        container_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[container_items_indexes_key].results
        backup_tags = {}
        rollback_tags = {}
        self._details = {"backup_tags": backup_tags, "rollback_tags": rollback_tags}

        for index in indexes:
            item = push_items[index]
            for repo, tags in six.iteritems(item.metadata.get("tags", {})):
                for tag in tags:
                    backup_tags.setdefault(repo, {})[tag] = None
                    rollback_tags.setdefault(repo, {})[tag] = None

    def _update_details(self, details):
        backup_tags = details
        self._update_tag_backups(backup_tags)

    @log_jsonl("BuildBackupMapping")
    def _run(self, on_update=None):
        push_items = self.external_resources["push_items"]
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        container_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[container_items_indexes_key].results
        for index in indexes:
            item = push_items[index]
            for repo, tags in six.iteritems(item.metadata.get("tags", {})):
                log_info("Marking backup for %s" % str(item))
                self.update_details({"repo": repo, "tags": tags})
                on_update()
        self.results.results = self._details


class StepPushContainerImgs(Step):
    """Push Container items to destination.

    Expected step args:
        (container_items_indexes_key, target_settings)
    Expected shared results:
        <container_items_indexes_key>: [ list of indexes of container push items in push_items]
    Produced details:
        {
            "items": { <repo>: { <tag>: { "state" : <state>, "source": <source> } }}
        }
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
        "push_items": list of push items
    Results:
        None
    """

    NAME = "StepPushContainerImgs"

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        items = {}
        self._details = {"items": items}
        container_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[container_items_indexes_key].results
        for index in indexes:
            item = push_items[index]
            for repo, tags in six.iteritems(item.metadata.get("tags", {})):
                items[repo] = {}
                for tag in tags:
                    items[repo][tag] = {
                        "state": "ready",
                        "source": item.metadata["pull_data"],
                    }

    def _push_container_item(self, item):
        raise NotImplementedError

    @log_jsonl("PushContainers")
    def _run(self, on_update=None):
        push_items = self.external_resources["push_items"]
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        container_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[container_items_indexes_key].results

        for index in indexes:
            item = push_items[index]
            log_info("Pushing %s", str(item))
            self._push_container_item(item)


class StepSignContainers(Step):
    """Sign pushed containers.

    Expected step args:
        (container_items_indexes_key)
    Expected step kwargs:
        {"container_signing_enabled": <True|False=default>}
    Expected shared results:
        <container_items_indexes_key>: [ list of indexes of container push items in push_items]
    Produced details:
        {
            "items": { <repo>: { <tag> { "state": ready|signed, "source": <pull_url>}}}
    Results:
        None
    """

    NAME = "StepSignContainers"

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        items = {}
        self._details = {"items": items}
        container_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[container_items_indexes_key].results

        for index in indexes:
            item = push_items[index]
            for repo, tags in six.iteritems(item.metadata.get("tags", {})):
                items[repo] = {}
                for tag in tags:
                    items[repo][tag] = {
                        "state": "ready",
                        "source": item.metadata["pull_data"],
                    }

    def _pre_run(self):
        if not self.step_kwargs.get("container_signing_enabled"):
            self.skip = True
            self.skip_reason = "Container signing for the target is not enabled"

    def _sign_container_items(self, sign_metadata):
        raise NotImplementedError

    @log_jsonl("SignContainers")
    def _run(self, on_update=None):
        push_items = self.external_resources["push_items"]
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        container_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[container_items_indexes_key].results

        sign_metadata = []
        for index in indexes:
            item = push_items[index]
            for repo, tags in six.iteritems(item.metadata.get("tags", {})):
                sign_metadata.append((repo, tags, item.metadata["pull_data"]))
                log_info("Signing %s %s", repo, tags)
        self._sign_container_items(sign_metadata)


class StepPushOperators(Step):
    """Push container operators.

    Expected step args:
        (operators_items_indexes_key,)
    Expected step kwargs:
        {"autoupload_operators": <True=default|False>,
         "docker_reference_registry": <str>,
         "iib_server":<str>}
    Expected shared results:
        <operator_items_indexes_key>: [ list of indexes of operator push items in push_items]
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
        "hub": pubhub xmlrpcserver instance
        "task_id": task id
        "target_name": target name
    Produced details:
        {
            "items": { <bundle_ref> { "state": ready|error|pushed}}
        }
    Results:
        None
    """

    NAME = "StepPushOperators"

    @staticmethod
    def _get_immutable_tag(item):
        """Return immutable tag from operator push item production tags."""
        if item.metadata["v_r"] in list(item.metadata["tags"].values())[0]:
            return item.metadata["v_r"]
        # if v_r tag is not in destination tags, return tag with the most numbers in it
        # This code is usually trigerred when push is not initiated from ET but manually
        # by user
        tags = []
        for tag in list(item.metadata["tags"].values())[0]:
            tags.append((len(re.split(r"\d+", tag)), tag))
        return sorted(tags)[-1][1]

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        items = {}
        self._details = {"items": items}
        for item in push_items:
            bundle_repo = list(item.metadata["destination"]["tags"].keys())[0]
            bundle_ref = "{0}/{1}:{2}".format(
                self.step_kwargs["docker_reference_registry"],
                bundle_repo,
                self._get_immutable_tag(item),
            )
            items[bundle_ref] = "ready"

    def _pre_run(self):
        if not self.step_kwargs.get("auto_upload_operators", True):
            self.skip = True
            self.skip_reason = "Automatic uploading of operators is not enabled"
        if not self.step_kwargs.get("iib_server"):
            self.skip = True
            self.skip_reason = "IIB server is not defined"

    def _push_operator_items(self, items, osev):
        raise NotImplementedError

    def pyxis_get_ocp_versions(self, item):
        """
        Get a list of supported ocp versions from Pyxis.

        Args:
            push_item: (_PushItem)
                Push item for which the OCP version range will be found out.

        Returns:
            list(str): list of supported OCP versions as returned by Pyxis.
        """
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        ocp_versions = item.metadata["com.redhat.openshift.versions"]
        log_info("Getting OCP versions of '{0}' from Pyxis.".format(ocp_versions))
        entry_point = (
            "pubtools-pyxis",
            "console_scripts",
            "pubtools-pyxis-get-operator-indices",
        )
        entry_point_fn = pkg_resources.load_entry_point(*entry_point)
        args = [
            "cmd",
            "--pyxis-server",
            self.step_kwargs["pyxis_server"],
            "--pyxis-krb-principal",
            self.step_kwargs["pyxis_krb_principal"],
            "--pyxis-krb-ktfile",
            self.step_kwargs["pyxis_krb_ktfile"],
            "--pyxis-ssl-crtfile",
            self.step_kwargs["pyxis_ssl_crtfile"],
            "--pyxis-ssl-keyfile",
            self.step_kwargs["pyxis_ssl_keyfile"],
            "--ocp-versions-range",
            ocp_versions,
        ]
        if self.step_kwargs.get("pyxis_insecure"):
            args += ["--pyxis-insecure"]
        data = entry_point_fn(args)
        if not data:
            msg = "Pyxis has returned no OCP versions for '{0}' specified in build {1}.".format(
                ocp_versions, item.metadata["build"]["build_id"]
            )
            raise ValueError(msg)
        # Versions returned by Pyxis don't contain 'v' at the front (4.5 -> v4.5)
        return ["v{0}".format(item["ocp_version"]) for item in data]

    @log_jsonl("PushOperators")
    def _run(self, on_update=None):
        push_items = self.external_resources["push_items"]
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        operator_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[operator_items_indexes_key].results
        iib_by_ose = {}
        ocp_versions_resolved = {}

        for index in indexes:
            item = push_items[index]
            ocp_versions = item.metadata["com.redhat.openshift.versions"]
            if ocp_versions not in ocp_versions_resolved:
                ocp_versions_resolved[ocp_versions] = self.pyxis_get_ocp_versions(item)
            for version in ocp_versions_resolved[ocp_versions]:
                iib_by_ose.setdefault(version, []).append(item)
        for osev, items in iib_by_ose.items():
            log_info("Pushing operators for %s: %s", osev, [str(x) for x in items])
            self._push_operator_items(items, osev)


class StepMergeManifestList(Step):
    """Merge manifest lists of pushed containers.

    Expected step args:
        (container_items_indexes_key, target_settings)
    Produced details:
        {
            "items": { <repo>: { <tag> { "state": ready|error|pushed,
                                         "source": <pull_url>}}}
        }
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
        "hub": pubhub xmlrpcserver instance
        "task_id": task id
    Results:
        None
    """

    NAME = "StepMergeManifestList"

    def _init_details(self):
        push_items = self.external_resources["push_items"]
        items = {}
        self._details = {"items": items}
        for item in push_items:
            for repo, tags in six.iteritems(item.metadata.get("tags", {})):
                items[repo] = {}
                for tag in tags:
                    items[repo][tag] = {
                        "state": "ready",
                        "source": item.metadata["pull_data"],
                    }

    def _merge_manifest_list(self, item):
        raise NotImplementedError

    @log_jsonl("MergeManifestList")
    def _run(self, on_update=None):
        push_items = self.external_resources["push_items"]
        operator_items_indexes_key = self.step_args[0]
        indexes = self._shared_results[operator_items_indexes_key].results
        for index in indexes:
            item = push_items[index]
            self._merge_manifest_list(item)


class StepRollback(Step):
    """Rollback step to revert data if sequence fails.

    Expected step args:
        (backup_results_key,)
    Produced details:
        {
            "items": { <repo>: { <tag>: (ready,removed,replaced) }}
        }
    Expected external resources:
        "log_debug": log debug callback or None
        "log_info": log info callback or None
        "log_warning": log warning callback or None
        "log_error": log error callback or None
    Results:
        None
    """

    NAME = "StepRollback"

    def _init_details(self):
        backup_results_key = self.step_args[0]
        items = {}
        self._details = {"items": items}
        backup_results = self._shared_results[backup_results_key].results

        for repo, tags in backup_results.get("backup_tags", {}).items():
            items[repo] = {}
            for tag in tags:
                items[repo][tag] = "ready"
        for repo, tags in backup_results.get("rollback_tags", {}).items():
            items[repo] = {}
            for tag in tags:
                items[repo][tag] = "ready"

    def _update_details(self, details):
        repo, tag, state = details
        self._details["items"][repo][tag] = state

    def _rollback(self, repo, to_rollback):
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        log_info("Removing", repo, to_rollback)
        raise NotImplementedError

    def _restore(self, repo, to_restore):
        log_info = self.external_resources.get("log_info", lambda *args, **kwargs: ())
        log_info("Restoring", repo, to_restore)
        raise NotImplementedError

    @log_jsonl("Rollback")
    def _run(self, on_update=None):
        backup_results_key = self.step_args[0]
        rollback = self._shared_results[backup_results_key].results.get("rollback_tags", {})
        backup = self._shared_results[backup_results_key].results.get("backup_tags", {})
        for repo, tags in rollback.items():
            for to_rollback in tags:
                self._rollback(repo, to_rollback)
                self.update_details((repo, to_rollback, "removed"))
        for repo, tags in backup.items():
            for to_restore in tags:
                self._restore(repo, to_restore)
                self.update_details((repo, to_restore, "restored"))



import argparse
import contextlib
import functools
import json
import logging
import os
import pkg_resources
import sys
import textwrap

from six import StringIO
from pubtools.pluggy import pm

LOG = logging.getLogger("pubtools.quay")

INTERNAL_DELIMITER = "----"


def setup_arg_parser(args):
    """
    Set up ArgumentParser with the provided arguments.

    Args:
        args (dict)
            Dictionary of argument aliases and options to be consumed by ArgumentParser.
    Returns:
        (ArgumentParser) Configured instance of ArgumentParser.
    """
    parser = argparse.ArgumentParser()
    arg_groups = {}
    for aliases, arg_data in args.items():
        holder = parser
        if "group" in arg_data:
            arg_groups.setdefault(arg_data["group"], parser.add_argument_group(arg_data["group"]))
            holder = arg_groups[arg_data["group"]]
        action = arg_data.get("action")
        if not action and arg_data["type"] == bool:
            action = "store_true"
        kwargs = {
            "help": arg_data.get("help"),
            "required": arg_data.get("required", False),
            "default": arg_data.get("default"),
        }
        if action:
            kwargs["action"] = action
        else:
            kwargs["type"] = arg_data.get("type", "str")
            kwargs["nargs"] = arg_data.get("count")

        holder.add_argument(*aliases, **kwargs)

    return parser


def add_args_env_variables(parsed_args, args):
    """
    Add argument values from environment variables.

    Args:
        parsed_args ():
            Parsed arguments object.
        args (dict):
            Argument definition.
    Returns:
        Modified parsed arguments object.
    """
    for aliases, arg_data in args.items():
        named_alias = [x.lstrip("-").replace("-", "_") for x in aliases if x.startswith("--")][0]
        if arg_data.get("env_variable"):
            if not getattr(parsed_args, named_alias) and os.environ.get(arg_data["env_variable"]):
                setattr(parsed_args, named_alias, os.environ.get(arg_data["env_variable"]))
    return parsed_args


def send_umb_message(urls, props, cert, topic, body=None, client_key=None, ca_cert=None):
    """
    Send a UMB message.

    Args:
        urls ([str]):
            URLs to send the message to.
        props (dict):
            Message properties dictionary.
        cert (str):
            Path to certificate for SSL authentication.
        topic (str):
            Topic to send the message to.
        body (str):
            Body of the message.
        client_key (str):
            Path to a private key for accessing the certificate.
        ca_cert (str):
            Path to CA certificate.
    """
    from rhmsg.activemq.producer import AMQProducer

    producer = AMQProducer(
        urls=urls,
        certificate=cert,
        private_key=client_key,
        topic=topic,
        trusted_certificates=ca_cert,
    )
    if not body:
        body = json.dumps(props).encode("utf-8")
    producer.send_msg(props, body)


@contextlib.contextmanager
def capture_stdout():
    """Capture sys.stdout to stream buffer."""
    old_stdout = sys.stdout
    sys.stdout = new_stdout = StringIO()

    try:
        yield new_stdout
    finally:
        sys.stdout = old_stdout


@contextlib.contextmanager
def setup_entry_point_cli(entry_tuple, name, args, environ_vars):
    """
    Set up an entrypoint as a context manager.

    Args:
        entry_tuple ((str, str, str)):
            Tuple consisting of dependency, category, and entrypoint.
        name: (str):
            Entrypoint name.
        args ([str]):
            Entrypoint arguments.
        environ_vars (dict):
            Env variable names and values to set for the entrypoint.
    """
    orig_argv = sys.argv[:]
    orig_environ = os.environ.copy()

    try:
        # First argv element is always the entry point name.
        # For a console_scripts entry point, this will be the same value
        # as if the script was invoked directly. For any other kind of entry point,
        # this value is probably meaningless.
        sys.argv = [name]
        sys.argv.extend(args)
        for key in environ_vars:
            os.environ[key] = environ_vars[key]
        entry_point_func = pkg_resources.load_entry_point(*entry_tuple)
        yield entry_point_func
    finally:
        sys.argv = orig_argv[:]
        os.environ.update(orig_environ)

        to_delete = [key for key in os.environ if key not in orig_environ]
        for key in to_delete:
            del os.environ[key]


def run_entrypoint(entry_tuple, name, args, environ_vars):
    """
    Run an entrypoint function and return its return value.

    Args:
        entry_tuple ((str, str, str)):
            Tuple consisting of dependency, category, and entrypoint.
        name: (str):
            Entrypoint name.
        args ([str]):
            Entrypoint arguments.
        environ_vars (dict):
            Env variable names and values to set for the entrypoint.

    Returns (str):
        Data returned by the entrypoint.
    """
    raw_args = " ".join([entry_tuple[2]] + args)
    wrapped_lines = textwrap.wrap(
        raw_args, 100, subsequent_indent="  ", break_on_hyphens=False, break_long_words=False
    )

    LOG.info("Running task with arguments:")
    for idx, line in enumerate(wrapped_lines):
        suffix = ""
        if idx != len(wrapped_lines) - 1:
            # shell-style backslash to indicate continuation
            suffix = " \\"
        LOG.info("%s%s", line, suffix)

    with setup_entry_point_cli(entry_tuple, name, args, environ_vars) as entry_func:
        with capture_stdout():
            pyret = entry_func()

    return pyret


def get_internal_container_repo_name(external_name):
    """
    Transform a repository name to an internal form in which it exists on Quay.io.

    Expected input format: <namespace>/<product>
    Generated output format: <namespace>----<product>

    NOTE: Repositories without a delimeter "/" may actually exist. In that case, the function
    simply returns the repo without any alterations.

    Args:
        external_name (str):
            External repository name.
    Returns:
        Internal repository name.
    """
    if external_name.count("/") == 0:
        return external_name

    if external_name.count("/") > 1 or external_name[0] == "/" or external_name[-1] == "/":
        raise ValueError(
            "Input repository containing a delimeter should "
            "have the format '<namespace>/<product>'",
            external_name,
        )

    return external_name.replace("/", INTERNAL_DELIMITER)


def get_external_container_repo_name(internal_name):
    """
    Transform a repository name to an external form in which it's visible to customers.

    Expected input format: <namespace>----<product>
    Generated output format: <namespace>/<product>

    NOTE: Repositories without a delimeter "----" may actually exist. In that case, the function
    simply returns the repo without any alterations.

    Args:
        internal_name (str):
            Internal repository name.
    Returns:
        External repository name.
    """
    if internal_name.count(INTERNAL_DELIMITER) == 0:
        return internal_name

    if (
        internal_name.count(INTERNAL_DELIMITER) > 1
        or internal_name.find(INTERNAL_DELIMITER) == 0
        or internal_name.find(INTERNAL_DELIMITER) == len(internal_name) - 4
    ):
        raise ValueError(
            "Input repository containing a delimeter should "
            "have the format '<namespace>----<product>'",
            internal_name,
        )

    return internal_name.replace(INTERNAL_DELIMITER, "/")


def task_status(event):
    """Helper function. Expand as necessary."""  # noqa: D401
    return dict(event={"type": event})


def log_step(step_name):
    """
    Log status for methods which constitute an entire task step.

    Args:
        step_name (str):
            Name of the task step, e.g., "Tag images".
    """
    event_name = step_name.lower().replace(" ", "-")

    def decorate(fn):
        @functools.wraps(fn)
        def fn_wrapper(*args, **kwargs):
            try:
                LOG.info("%s: Started", step_name, extra=task_status("%s-start" % event_name))
                ret = fn(*args, **kwargs)
                LOG.info("%s: Finished", step_name, extra=task_status("%s-end" % event_name))
                return ret
            except Exception:
                LOG.error("%s: Failed", step_name, extra=task_status("%s-error" % event_name))
                raise

        return fn_wrapper

    return decorate


def get_pyxis_ssl_paths(target_settings):
    """
    Get certificate and key paths for Pyxis SSL authentication.

    First attempt is made by invoking the hook implementation 'get_cert_key_paths'.
    If nothing is returned (no hook implementation is registered), fallback on target settings
    values of 'pyxis_ssl_cert' and 'pyxis_ssl_key'. If multiple values are returned (multiple
    hook implementations are registered), take the first response (from the implementation which
    was registered last). Otherwise, raise an error.

    Args:
        target_settings (dict):
            Dictionary containing various task-related settings.
    Returns ((str, str)):
        Paths to Pyxis SSL certificate and key.
    """
    result = pm.hook.get_cert_key_paths(server_url=target_settings["pyxis_server"])
    if result:
        cert, key = result
    elif "pyxis_ssl_cert" not in target_settings or "pyxis_ssl_key" not in target_settings:
        raise ValueError(
            "No key and certificate paths were provided for Pyxis SSL authentication. "
            "Please, either provide hook implementation for 'get_cert_key_paths_plugin' or "
            "specify 'pyxis_ssl_cert' and 'pyxis_ssl_key' in the target settings."
        )
    else:
        cert = target_settings["pyxis_ssl_cert"]
        key = target_settings["pyxis_ssl_key"]

    return (cert, key)



import logging

import requests

from .exceptions import (
    BadPushItem,
    ManifestTypeError,
)
from .utils.misc import (
    get_internal_container_repo_name,
    log_step,
)
from .quay_client import QuayClient
from .tag_images import tag_images
from .manifest_list_merger import ManifestListMerger

LOG = logging.getLogger("pubtools.quay")


class ContainerImagePusher:
    """
    Push container images to Quay.

    No validation is performed, push items are expected to be correct.
    """

    def __init__(self, push_items, target_settings):
        """
        Initialize.

        Args:
            push_items ([ContainerPushItem]):
                List of push items.
            target_settings (dict):
                Target settings.
        """
        self.push_items = push_items
        self.target_settings = target_settings

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")
        self._src_quay_client = None
        self._dest_quay_client = None

    @property
    def src_quay_client(self):
        """Create and access QuayClient for source image."""
        if self._src_quay_client is None:
            self._src_quay_client = QuayClient(
                self.target_settings["source_quay_user"],
                self.target_settings["source_quay_password"],
                self.quay_host,
            )
        return self._src_quay_client

    @property
    def dest_quay_client(self):
        """Create and access QuayClient for dest image."""
        if self._dest_quay_client is None:
            self._dest_quay_client = QuayClient(
                self.target_settings["dest_quay_user"],
                self.target_settings["dest_quay_password"],
                self.quay_host,
            )
        return self._dest_quay_client

    @classmethod
    def run_tag_images(cls, source_ref, dest_refs, all_arch, target_settings):
        """
        Prepare the "tag images" entrypoint with all the necessary arguments and run it.

        Args:
            source_ref (str):
                Source image reference.
            dest_refs ([str]):
                List of destination references.
            all_arch (bool):
                Whether all architectures should be copied.
            target_settings (dict):
                Settings used for setting the values of the function parameters.
        """
        # TODO: do we want to do some registry-proxy -> quay transformation?
        # TODO: tag-images only supports quay.io hostname, should we extend the functionality?
        # TODO: should this command always be performed remotely?
        tag_images(
            source_ref,
            dest_refs,
            all_arch=all_arch,
            quay_user=target_settings["dest_quay_user"],
            quay_password=target_settings["dest_quay_password"],
            container_exec=True,
            container_image=target_settings["skopeo_image"],
            docker_url=target_settings.get("docker_host") or "unix://var/run/docker.sock",
            docker_timeout=target_settings.get("docker_timeout"),
            docker_verify_tls=target_settings.get("docker_tls_verify") or False,
            docker_cert_path=target_settings.get("docker_cert_path") or None,
            send_umb_msg=True,
            umb_urls=target_settings["docker_settings"]["umb_urls"],
            umb_cert=target_settings["docker_settings"].get(
                "umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"
            ),
            # assumption that we'll continue using .pem format
            umb_client_key=target_settings["docker_settings"].get(
                "umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"
            ),
            umb_ca_cert=target_settings["docker_settings"].get(
                "umb_ca_cert", "/etc/pki/tls/certs/ca-bundle.crt"
            ),
        )

    def copy_source_push_item(self, push_item):
        """
        Perform the tagging operation for a push item containing a source image.

        Args:
            push_item (ContainerPushItem):
                Source container push item.
        """
        LOG.info("Copying push item '{0}' as a source image".format(push_item))
        source_ref = push_item.metadata["pull_url"]
        dest_refs = []
        image_schema = "{host}/{namespace}/{repo}:{tag}"
        namespace = self.target_settings["quay_namespace"]

        for repo, tags in sorted(push_item.metadata["tags"].items()):
            internal_repo = get_internal_container_repo_name(repo)
            for tag in tags:
                dest_ref = image_schema.format(
                    host=self.quay_host,
                    namespace=namespace,
                    repo=internal_repo,
                    tag=tag,
                )
                dest_refs.append(dest_ref)

        self.run_tag_images(source_ref, dest_refs, True, self.target_settings)

    def run_merge_workflow(self, source_ref, dest_refs):
        """
        Perform Docker push and manifest list merge workflow.

        The difference in this workflow is that all single arch images are first copied via
        digest, and then their respective manifest lists are merged.

        Args:
            source_ref (str):
                Source image reference.
            dest_refs ([str]):
                List of destination references which need manifest merging.
        """
        image_schema = "{repo}@{digest}"
        source_repo = source_ref.split(":")[0]

        # get unique destination repositories
        dest_repos = sorted(list(set([ref.split(":")[0] for ref in dest_refs])))
        source_ml = self.src_quay_client.get_manifest(source_ref, manifest_list=True)

        # copy each arch source image to all destination repos
        for manifest in source_ml["manifests"]:
            source_image = image_schema.format(repo=source_repo, digest=manifest["digest"])
            dest_images = [
                image_schema.format(repo=dest_repo, digest=manifest["digest"])
                for dest_repo in dest_repos
            ]
            self.run_tag_images(source_image, dest_images, False, self.target_settings)

        for dest_ref in dest_refs:
            LOG.info(
                "Merging manifest lists of source '{0}' and destination '{1}'".format(
                    source_ref, dest_ref
                )
            )
            merger = ManifestListMerger(source_ref, dest_ref, host=self.quay_host)
            merger.set_quay_clients(self.src_quay_client, self.dest_quay_client)
            merger.merge_manifest_lists()

    def copy_multiarch_push_item(self, push_item, source_ml):
        """
        Evaluate the correct tagging and manifest list merging strategy of multiarch push item.

        There are two workflows of multiarch images: Simple copying, or manifest list merging.
        Destination tags are sorted, and correct workflow is performed on them.

        Args:
            push_items (ContainerPushItem):
                Multiarch container push item.
            source_ml (dict):
                Manifest list of the source image.
        """
        LOG.info("Copying push item '{0}' as a multiarch image.".format(push_item))
        source_ref = push_item.metadata["pull_url"]
        simple_dest_refs = []
        merge_mls_dest_refs = []

        image_schema = "{host}/{namespace}/{repo}:{tag}"
        namespace = self.target_settings["quay_namespace"]

        for repo, tags in sorted(push_item.metadata["tags"].items()):
            internal_repo = get_internal_container_repo_name(repo)
            for tag in tags:
                dest_ref = image_schema.format(
                    host=self.quay_host,
                    namespace=namespace,
                    repo=internal_repo,
                    tag=tag,
                )
                try:
                    dest_ml = self.dest_quay_client.get_manifest(dest_ref, manifest_list=True)
                    LOG.info(
                        "Getting missing archs between images '{0}' and '{1}'".format(
                            source_ref, dest_ref
                        )
                    )
                    missing_archs = ManifestListMerger.get_missing_architectures(source_ml, dest_ml)
                    # Option 1: Destination doesn't contain extra archs, ML merging is unnecessary
                    if not missing_archs:
                        simple_dest_refs.append(dest_ref)
                    # Option 2: Destination has extra archs, MLs will be merged
                    else:
                        merge_mls_dest_refs.append(dest_ref)
                except requests.exceptions.HTTPError as e:
                    # Option 3: Destination tag doesn't exist, no ML merging
                    if e.response.status_code == 404 or e.response.status_code == 401:
                        simple_dest_refs.append(dest_ref)
                    else:
                        raise

        if simple_dest_refs:
            LOG.info(
                "Copying image {0} to {1} destinations without merging manifest lists".format(
                    source_ref, len(simple_dest_refs)
                )
            )
            self.run_tag_images(source_ref, simple_dest_refs, True, self.target_settings)
        if merge_mls_dest_refs:
            LOG.info(
                "Copying image {0} to {1} destinations and merging manifest lists".format(
                    source_ref, len(merge_mls_dest_refs)
                )
            )
            self.run_merge_workflow(source_ref, merge_mls_dest_refs)

    @log_step("Push images to Quay")
    def push_container_images(self):
        """
        Push container images to Quay.

        Two image types are supported: source images and multiarch images. Non-source, single arch
        images are not supported. In case of multiarch images, manifest list merging is performed if
        destination image contains more architectures than source.
        """
        for item in self.push_items:
            try:
                source_ml = self.src_quay_client.get_manifest(
                    item.metadata["pull_url"], manifest_list=True
                )
            except ManifestTypeError:
                source_ml = None

            # this metadata field indicates a source image
            sources_for_nvr = (
                item.metadata["build"]
                .get("extra", {})
                .get("image", {})
                .get("sources_for_nvr", None)
            )
            if not sources_for_nvr and not source_ml:
                raise BadPushItem(
                    "Push item '{0}' contains a single-arch image that's not a "
                    "source image. This use-case is not supported".format(item)
                )
            # Source image
            if sources_for_nvr:
                self.copy_source_push_item(item)
            # Multiarch images
            else:
                self.copy_multiarch_push_item(item, source_ml)



import abc
import copy
import datetime


def isodate_now():
    """Return current datetime in iso8601 format."""
    return "%s%s" % (datetime.datetime.utcnow().isoformat(), "Z")


class StepFailedError(Exception):
    """Exception indidating failure of a step."""


class Secret(object):
    """Class for storing sensitive values used as argument for Step class."""

    def __init__(self, val):
        """Init secret instance."""
        self.value = val

    def __str__(self):
        """Return *CENSORED* constant string."""
        return "*CENSORED*"


class StepState(object):
    """Enum-like class to store step state."""

    def __init__(self):
        """Init step state."""
        self._state = "ready"

    def get(self):
        """Return step state."""
        return self._state

    def set(self, state):
        """Set step state."""
        if state not in ("ready", "prep", "running", "finished", "failed", "error"):
            raise ValueError(
                "state has to be one of "
                "('ready', 'prep', 'running', "
                "'finished', 'failed', 'error')"
            )
        self._state = state


class StepResults(object):
    """Class to store results of step."""

    def __init__(self):
        """Init step results."""
        self.results = {}
        self.errors = {}

    def dump(self):
        """Return step results in json-compatible dict object."""
        return {"results": self.results, "errors": self.errors}

    def load(self, obj):
        """Load step results from dict object."""
        self.results = obj["results"]
        self.errors = obj["errors"]


class Step(object):
    """Base class for a Step.

    How to use this class: Few things are needed to implement custom step class.
    First, user needs to overwrite _run method which should do include all the code
    which is meant to be do desired step operation.
    In run method, user can access following instance attributes:
    `step_args`, `step_kwargs` - set when Step is initialized. These two variables
        are meant to hold data for the step. User needs to design step to work only
        with data which are json-compatible
    `shared_results` - shared dict-like object where step can store data for
        another steps or load data generated by previously ran steps.
    `external_resources` - resources which are needed for step to run but are not data.
        This can be for example logger, initialized client for external service and
        similar
    `results` - `StepResults` instance used to store data generated by the step.

    To provide detailed info about step status, step can store these details in _details
    attribute of the step instance. Details can contain anything json compatible. To
    set details to initial state, user needs to overwrite _init_details.
    Later, update_details method can be used to updating the details.
    User can overwrite `_pre_run` method to do any kind of 'lazy' preparation of data or
    set `skip` and `skip_reason` variables in the instance to prevent step from the
    execution.
    When there's data error or wrong data are provided, _run method can
    raise StepFailedError to set step to 'failed' state. Failed state indicates
    there's problem with data or configuration and step in this state cannot be
    executed again. If any other exception occurs and is not caught in _run, step is set
    to 'error' state and can be executed again. Only other two states indicating step is
    able to be executed are 'ready' and 'prep'. Ready state is set after initialization
    of the step instance. Step is switched to 'prep' state just before `_pre_run`
    is called. If step is in 'prep' state  `_pre_run` is not called again. After step
    finished the execution, and if there wasn't any error, `results` instance attribute
    is stored to `shared_results`.
    Last thing to do is to set NAME class attribute identify type of step
    """

    NAME = "AbstractSequenceStep"
    __metaclass__ = abc.ABCMeta

    def __init__(self, uid, step_args, step_kwargs, shared_results, external_resources=None):
        """Initilize the step.

        Args:
            uid: (str)
                An unique id for indentifying two steps of the same class
            step_args: (list)
                Arguments used for the step
            step_kwargs: (dict)
                Dictionary of a name arguments for the step
            shared_results: (dict-like object)
                Object to store shared data between steps
            external_resources: any-object
                Reference for external resources (in any form) which are constant and
                shouldn't be part of the step state or step data
        """
        self.step_args = step_args

        self.masked_args = [str(arg) if isinstance(arg, Secret) else arg for arg in step_args]
        self.masked_kwargs = dict(
            [
                (k, str(arg)) if isinstance(arg, Secret) else (k, arg)
                for k, arg in step_kwargs.items()
            ]
        )

        self.step_args = [arg.value if isinstance(arg, Secret) else arg for arg in step_args]
        self.step_kwargs = dict(
            [
                (k, arg.value) if isinstance(arg, Secret) else (k, arg)
                for k, arg in step_kwargs.items()
            ]
        )

        self._shared_results = shared_results
        self.external_resources = external_resources or {}
        self.uid = uid

        self.stats = {}
        self._details = {}
        self._details_initted = False
        self.skip = False
        self.skip_reason = ""
        self._state = StepState()
        self.results = StepResults()

    @property
    def fullname(self):
        """Full name of class instance."""
        return "%s:%s" % (self.NAME, self.uid)

    @property
    def state(self):
        """Property for retrieving class state."""
        return self._state.get()

    def set_state(self, state):
        """Set step state."""
        self._state.set(state)

    def run(self, on_update=None):
        """Run the step code.

        Step is expected to run when step state is ready, prep or error. For other
        state running the step code is omitted. If step is in ready state,
        _pre_run method is executed first and state is switched to prep.
        After prep phase finishes, and skip is not set to True, _run method containing
        all the code for running the step is executed.
        After _run finishes, step state is set to failed, error or finished. Statistics
        of step are update and potential results of step are stored in shared data object
        After every change of step state, on_update callback is called if set
        """
        _on_update = on_update if on_update else lambda: ()
        self._reset_stats()
        if self.state == "ready":
            self.stats["started"] = isodate_now()
            self.set_state("prep")
            self._pre_run()
            _on_update()
        try:
            if self.state not in ("prep", "error"):
                return
            if not self.skip:
                self.set_state("running")
                _on_update()
                self._run(on_update=_on_update)
        except StepFailedError:
            self.set_state("failed")
        except Exception:
            self.set_state("error")
            raise
        else:
            self.set_state("finished")
        finally:
            self._finish_stats()
            self._store_results()
            _on_update()

    def _pre_run(self):
        """Execute code needed before step run.

        In this method, all neccesary preparation of data can be done.
        It can be also used to determine if step should run or not by setting
        self.skip to True and providing self.skip_reason string with explanation.
        """
        pass

    def _reset_stats(self):
        self.stats = {
            "started": None,
            "finished": None,
            "skip": self.skip,
            "skip_reason": self.skip_reason,
            "skipped": None,
            "state": self.state,
        }

    def _finish_stats(self):
        self.stats["finished"] = isodate_now()
        self.stats["skipped"] = self.skip
        self.stats["skip"] = self.skip
        self.stats["skip_reason"] = self.skip_reason
        self.stats["state"] = self.state

    def _store_results(self):
        self._shared_results[self.fullname] = self.results

    @abc.abstractmethod
    def _run(self, on_update=None):  # pragma: no cover
        """Run code of the step.

        Method expects raise StepFailedError if step code fails due data error
        (incorrect configuration or missing/wrong data). That ends with step
        state set to failed.
        If error occurs due to uncaught exception in this method, step state
        will be set to error
        """
        raise NotImplementedError

    def _init_details(self):
        """Initialize step details.

        It's called internally before step details exported to json and before
        calling update_details if hasn't been called before.
        """
        pass

    def _update_details(self, details):  # pragma: no cover
        """Update step details (needs to be overwritten)."""
        pass

    def update_details(self, details_args):
        """Update step details, init them if needed.

        Also init details if those haven't been initilized before
        """
        if not self._details_initted:
            self._init_details()
            self._details_initted = True
        self._update_details(details_args)

    def dump(self):
        """Dump step data into json compatible complex dictionary."""
        if not self._details_initted:
            self._init_details()
            self._details_initted = True
        if not self.stats:
            self._reset_stats()

        return {
            "name": self.NAME,
            "step_args": self.masked_args,
            "step_kwargs": self.masked_kwargs,
            "uid": self.uid,
            "details": copy.deepcopy(self._details),
            "stats": self.stats.copy(),
            "results": self.results.dump(),
        }

    @classmethod
    def load(cls, obj, shared_results):
        """Load step data from dictionary produced by dump method."""
        ret = cls(obj["uid"], obj["step_args"], obj["step_kwargs"], shared_results)
        ret._details = obj["details"]
        ret.stats = obj["stats"]
        ret.skip = obj["stats"]["skip"]
        ret.results.load(obj["results"])
        ret._details_initted = True
        ret.skip_reason = obj["stats"]["skip_reason"]
        ret.set_state(obj["stats"]["state"])
        return ret


class Stepper(object):
    """Class which runs sequence of steps."""

    def __init__(self, step_map):
        """Initialize the stepper.

        Args:
            step_map: (mapping of "step-name": <step_class>)
                Mapping of step names to step classes. Used when loading stepper from
                json-compatible dict data
        """
        self.step_map = step_map
        self.steps = []
        self.shared_results = {}
        self.current_step = None

    def add_step(self, step):
        """Add step to step sequence."""
        self.steps.append(step)

    def dump(self):
        """Dump stepper state and shared_results to json compatible dict."""
        steps = []
        out = {"steps": steps, "shared_results": self.shared_results}
        for step in self.steps:
            steps.append(step.dump())
        return out

    def load(self, stepper_obj):
        """Load and initialize stepper from data produced by dump method."""
        self.steps = []
        self.shared_results = stepper_obj["shared_results"]
        for step_obj in stepper_obj["steps"]:
            step = self.step_map[step_obj["name"]].load(step_obj, self.shared_results)
            self.steps.append(step)

    def run(self, start_from=0, on_error=None, on_update=None):
        """Run the stepper sequence."""
        _on_error = on_error if on_error else lambda s: ()
        try:
            for step in self.steps[start_from:]:
                self.current_step = step
                step.run(on_update=on_update)
        except Exception:
            _on_error(self)
            raise



import functools
import logging

from pubtools.pluggy import pm, task_context

from .utils.misc import setup_arg_parser, add_args_env_variables, send_umb_message
from .command_executor import LocalExecutor, RemoteExecutor, ContainerExecutor

LOG = logging.getLogger("pubtools.quay")

TAG_IMAGES_ARGS = {
    ("--source-ref",): {
        "help": "Source image reference.",
        "required": True,
        "type": str,
    },
    ("--dest-ref",): {
        "help": "Destination image reference. Multiple can be specified.",
        "required": True,
        "type": str,
        "action": "append",
    },
    ("--all-arch",): {
        "help": "Flag of whether to copy all architectures of an image (if multiatch image)",
        "required": False,
        "type": bool,
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": False,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
    ("--remote-exec",): {
        "help": "Flag of whether the commands should be executed on a remote server.",
        "required": False,
        "type": bool,
    },
    ("--ssh-remote-host",): {
        "help": "Hostname for remote execution.",
        "required": False,
        "type": str,
    },
    ("--ssh-remote-host-port",): {
        "help": "Port of the remote host",
        "required": False,
        "type": int,
    },
    ("--ssh-reject-unknown-host",): {
        "help": "Flag of whether to reject an SSH host when it's not found among known hosts.",
        "required": False,
        "type": bool,
    },
    ("--ssh-username",): {
        "help": "Username for SSH connection. Defaults to local username.",
        "required": False,
        "type": str,
    },
    ("--ssh-password",): {
        "help": "Password for SSH. Will only be used if key-based validation is not available. "
        "Can be specified by env variable SSH_PASSWORD",
        "required": False,
        "type": str,
        "env_variable": "SSH_PASSWORD",
    },
    ("--ssh-key-filename",): {
        "help": "Path to the private key file for SSH authentication.",
        "required": False,
        "type": str,
    },
    ("--container-exec",): {
        "help": "Whether to execute the commands in a Docker container.",
        "required": False,
        "type": bool,
    },
    ("--container-image",): {
        "help": "Path to the container image in which to execute the commands. Must be "
        "downloadable without extra permissions.",
        "required": False,
        "type": str,
    },
    ("--docker-url",): {
        "help": "URL of the docker client that should run the container. Local socket by default.",
        "required": False,
        "type": str,
        "default": "unix://var/run/docker.sock",
    },
    ("--docker-timeout",): {
        "help": "Timeout for executing Docker commands. Disabled by default.",
        "required": False,
        "type": str,
    },
    ("--docker-verify-tls",): {
        "help": "Whether to perform TLS verification with the Docker client. Disabled by default.",
        "required": False,
        "type": bool,
    },
    ("--docker-cert-path",): {
        "help": "Path to Docker certificates for TLS authentication. '~/.docker' by default.",
        "required": False,
        "type": str,
    },
    ("--send-umb-msg",): {
        "help": "Flag of whether to send a UMB message",
        "required": False,
        "type": bool,
    },
    ("--umb-url",): {
        "help": "UMB URL. More than one can be specified.",
        "required": False,
        "type": str,
        "action": "append",
    },
    ("--umb-cert",): {
        "help": "Path to the UMB certificate for SSL authentication.",
        "required": False,
        "type": str,
    },
    ("--umb-client-key",): {
        "help": "Path to the UMB private key for accessing the certificate.",
        "required": False,
        "type": str,
    },
    ("--umb-ca-cert",): {
        "help": "Path to the UMB CA certificate.",
        "required": False,
        "type": str,
    },
    ("--umb-topic",): {
        "help": "UMB topic to send the message to.",
        "required": False,
        "type": str,
        "default": "VirtualTopic.eng.pub.quay_tag_image",
    },
}


def construct_kwargs(args):
    """
    Construct a kwargs dictionary based on the entered command line arguments.

    Args:
        args (argparse.Namespace):
            Parsed command line arguments.
    Returns (dict):
        Keyword arguments for the 'tag_images' function.
    """
    kwargs = args.__dict__

    # in args.__dict__ unspecified bool values have 'None' instead of 'False'
    for name, attributes in TAG_IMAGES_ARGS.items():
        if attributes["type"] is bool:
            bool_var = name[0].lstrip("-").replace("-", "_")
            if kwargs[bool_var] is None:
                kwargs[bool_var] = False

    # some exceptions have to be remapped
    kwargs["dest_refs"] = kwargs.pop("dest_ref")
    kwargs["umb_urls"] = kwargs.pop("umb_url")

    return kwargs


def tag_images(
    source_ref,
    dest_refs,
    all_arch=False,
    quay_user=None,
    quay_password=None,
    remote_exec=False,
    ssh_remote_host=None,
    ssh_remote_host_port=None,
    ssh_reject_unknown_host=False,
    ssh_username=None,
    ssh_password=None,
    ssh_key_filename=None,
    container_exec=False,
    container_image=None,
    docker_url="unix://var/run/docker.sock",
    docker_timeout=None,
    docker_verify_tls=False,
    docker_cert_path=None,
    send_umb_msg=False,
    umb_urls=[],
    umb_cert=None,
    umb_client_key=None,
    umb_ca_cert=None,
    umb_topic="VirtualTopic.eng.pub.quay_tag_image",
):
    """
    Tag images in Quay.

    Args:
        source_ref (str):
            Source image reference.
        dest_refs ([str]):
            List of destination image references.
        all_arch (bool):
            Whether to copy all architectures.
        quay_user (str):
            Quay username for Docker HTTP API.
        quay_password (str):
            Quay password for Docker HTTP API.
        remote_exec (bool):
            Whether to execute the command remotely. Takes precedence over container_exec.
        ssh_remote_host (str):
            Hostname for remote execution.
        ssh_remote_host_port (str):
            Port of the remote host.
        ssh_reject_unknown_host (bool):
            whether to reject an SSH host when it's not found among known hosts.
        ssh_username (str):
            Username for SSH connection. Defaults to local username.
        ssh_password (str):
            Password for SSH. Will only be used if key-based validation is not available.
        ssh_key_filename (str):
            Path to the private key file for SSH authentication.
        container_exec (bool):
            Whether to execute the commands in a Docker container.
        container_image (str):
            Path to the container image in which to execute the commands. Must be downloadable
            without extra permissions.
        docker_url (str):
            URL of the docker client that should run the container. Local socket by default.
        docker_timeout (int):
            Timeout for executing Docker commands. Disabled by default.
        docker_verify_tls (bool):
            Whether to perform TLS verification with the Docker client. Disabled by default.
        docker_cert_path (str):
            Path to Docker certificates for TLS authentication. '~/.docker' by default.
        send_umb_msg (bool):
            Whether to send UMB messages about the untagged images.
        umb_urls ([str]):
            AMQP broker URLs to connect to.
        umb_cert (str):
            Path to a certificate used for UMB authentication.
        umb_client_key (str):
            Path to a client key to decrypt the certificate (if necessary).
        umb_ca_cert (str):
            Path to a CA certificate (for mutual authentication).
        umb_topic (str):
            Topic to send the UMB messages to.
    """
    verify_tag_images_args(
        quay_user,
        quay_password,
        remote_exec,
        ssh_remote_host,
        container_exec,
        container_image,
        send_umb_msg,
        umb_urls,
        umb_cert,
    )

    if remote_exec:
        accept_host = not ssh_reject_unknown_host if ssh_reject_unknown_host else True
        executor_class = functools.partial(
            RemoteExecutor,
            ssh_remote_host,
            ssh_username,
            ssh_key_filename,
            ssh_password,
            ssh_remote_host_port,
            accept_host,
        )
    elif container_exec:
        if isinstance(docker_timeout, str):
            docker_timeout = int(docker_timeout)
        executor_class = functools.partial(
            ContainerExecutor,
            container_image,
            docker_url,
            docker_timeout,
            docker_verify_tls,
            docker_cert_path,
        )
    else:
        executor_class = functools.partial(LocalExecutor)

    with executor_class() as executor:
        executor.skopeo_login(quay_user, quay_password)
        executor.tag_images(source_ref, dest_refs, all_arch)

    pm.hook.quay_images_tagged(source_ref=source_ref, dest_refs=sorted(dest_refs))

    if send_umb_msg:
        props = {"source_ref": source_ref, "dest_refs": dest_refs}
        send_umb_message(
            umb_urls,
            props,
            umb_cert,
            umb_topic,
            client_key=umb_client_key,
            ca_cert=umb_ca_cert,
        )


def verify_tag_images_args(
    quay_user,
    quay_password,
    remote_exec,
    ssh_remote_host,
    container_exec,
    container_image,
    send_umb_msg,
    umb_urls,
    umb_cert,
):
    """Verify the presence of input parameters."""
    if remote_exec:
        if not ssh_remote_host:
            raise ValueError("Remote host is missing when remote execution was specified.")

    if container_exec:
        if not container_image:
            raise ValueError("Container image is missing when container execution was specified.")

    if (quay_user and not quay_password) or (quay_password and not quay_user):
        raise ValueError("Both user and password must be present when attempting to log in.")

    if send_umb_msg:
        if not umb_urls:
            raise ValueError("UMB URL must be specified if sending a UMB message was requested.")
        if not umb_cert:
            raise ValueError(
                "A path to a client certificate must be provided when sending a UMB message."
            )


def setup_args():
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(TAG_IMAGES_ARGS)


def tag_images_main(sysargs=None):
    """Entrypoint for image tagging."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, TAG_IMAGES_ARGS)

    kwargs = construct_kwargs(args)
    with task_context():
        tag_images(**kwargs)



import contextlib
import mock
import pkg_resources
import re

# flake8: noqa: D200, D107, D102, D105


def sort_dictionary_sortable_values(dictionary):
    """
    Sort lists which are values of a dictionary. May be used for better dict comparisons.

    Args:
        dictionary (dict):
            Dictionary whose values should be sorted.
    """
    for v in dictionary.values():
        if isinstance(v, list):
            v.sort()


def compare_logs(caplog, expected_logs):
    """
    Compare log messages between captured messages and expected ones.

    Expected messages can contain regexes.

    Args:
        caplog:
            Caplog generated by pytest.
        expected_logs ([str]):
            List of strings/regexes containing expected logs lines.
    """
    assert len(caplog.records) == len(expected_logs)

    for i in range(len(expected_logs)):
        log_line = caplog.records[i].getMessage()
        match = re.search(expected_logs[i], log_line)
        if match is None:
            raise AssertionError(
                "Captured log line '{0}' couldn't be matched with '{1}'".format(
                    log_line, expected_logs[i]
                )
            )


class FixedEntryPoint(object):
    """An EntryPoint which always points at the specified object rather than going
    through normal resolution."""

    def __init__(self, delegate, resolved_object):
        # These all work identically as in the real entry point
        self.name = delegate.name
        self.module_name = delegate.module_name
        self.attrs = delegate.attrs
        self.extras = delegate.extras
        self.dist = delegate.dist
        self.require = delegate.require
        self.__str__ = delegate.__str__

        # But any attempt to load or resolve will just return this rather than
        # the true underlying object
        self.load = lambda: resolved_object
        self.resolve = lambda: resolved_object


class EntryPointOverride(object):
    """Point a named entry point at a provided object temporarily."""

    def __init__(self, replacement, dist, group, name):
        self._name = name

        # Keep a reference to the original entry point
        dist = pkg_resources.get_distribution(dist)
        self._ep_map = dist.get_entry_map(group)
        self._old_ep = self._ep_map[self._name]

        # Prepare the replacement
        self._replace_ep = FixedEntryPoint(self._old_ep, replacement)

    def start(self):
        self._ep_map[self._name] = self._replace_ep

    def __enter__(self):
        self.start()
        return self

    def stop(self):
        self._ep_map[self._name] = self._old_ep

    def __exit__(self, exc_type, exc_value, traceback):
        self.stop()


@contextlib.contextmanager
def mock_entry_point(dist, group, name):
    """Point a given entry point at a new mock object for the duration of the current test."""
    new_mock = mock.Mock()
    override = EntryPointOverride(new_mock, dist, group, name)
    try:
        override.start()
        yield new_mock
    finally:
        override.stop()


class IIBRes:
    def __init__(self, index_image, index_image_resolved):
        self.index_image = index_image
        self.index_image_resolved = index_image_resolved



from copy import deepcopy
import logging

import requests

from .quay_client import QuayClient

LOG = logging.getLogger("pubtools.quay")


class ManifestListMerger:
    """Class containing logic for merging manifest lists of two images."""

    def __init__(
        self,
        src_image,
        dest_image,
        src_quay_username=None,
        src_quay_password=None,
        dest_quay_username=None,
        dest_quay_password=None,
        host=None,
    ):
        """
        Initialize.

        Args:
            src_image (str):
                Address to a new image whose manifest list contains the newer data.
            dest_image (str):
                Address to an older image whose data will be overwritten.
            src_quay_username (str):
                Quay username to get src_image. If ommited, external client instance should be set.
            src_quay_password (str):
                Quay password to get src_image. If ommited, external client instance should be set.
            dest_quay_username (str):
                Quay username to get dest_image. If ommited, external client instance should be set.
            dest_quay_password (str):
                Quay password to get dest_image. If ommited, external client instance should be set.
            host (str):
                Custom hostname to connect to. If ommited, standard quay.io will be used.
        """
        self.src_image = src_image
        self.dest_image = dest_image
        if src_quay_username and src_quay_password:
            self._src_quay_client = QuayClient(src_quay_username, src_quay_password, host)
        else:
            self._src_quay_client = None
        if dest_quay_username and dest_quay_password:
            self._dest_quay_client = QuayClient(dest_quay_username, dest_quay_password, host)
        else:
            self._dest_quay_client = None

    def set_quay_clients(self, src_quay_client, dest_quay_client):
        """
        Set client instances to be used for the HTTP API operations.

        Args:
            src_quay_client (QuayClient):
                Instance of QuayClient.
            dest_quay_client (QuayClient):
                Instance of QuayClient.
        """
        self._src_quay_client = src_quay_client
        self._dest_quay_client = dest_quay_client

    def merge_manifest_lists(self):
        """Merge manifest lists and upload to Quay. Main entrypoint method."""
        if not self._src_quay_client or not self._dest_quay_client:
            raise RuntimeError("QuayClient instance must be set for both source and dest images")

        LOG.info(
            "Merging manifest lists of images '{0}' and '{1}'".format(
                self.src_image, self.dest_image
            )
        )
        src_manifest_list = self._src_quay_client.get_manifest(self.src_image, manifest_list=True)
        dest_manifest_list = self._dest_quay_client.get_manifest(
            self.dest_image, manifest_list=True
        )

        missing_archs = self.get_missing_architectures(src_manifest_list, dest_manifest_list)
        new_manifest_list = self._add_missing_architectures(src_manifest_list, missing_archs)

        LOG.info("Uploading the new manifest list to '{0}'".format(self.dest_image))
        self._dest_quay_client.upload_manifest(new_manifest_list, self.dest_image)
        LOG.info("Merging manifests lists: complete.")

    @staticmethod
    def get_missing_architectures(src_manifest_list, dest_manifest_list):
        """
        Get architectures which are missing from the new source image.

        NOTE: this method assumes that images are built only for one OS. The following logic
        would need to be overwritten if multiple OS builds started to be made.

        Args:
            src_manifest_list (dict):
                Manifest list of the source image.
            dest_manifest_list (dict):
                Manifest list of the destination image.
        Returns ([dict]):
            List of arch manifest data present in destination image but missing from source.
        """
        missing_archs = []
        missing_archs_log = []
        src_archs = [arch["platform"]["architecture"] for arch in src_manifest_list["manifests"]]

        for dest_arch_dict in dest_manifest_list["manifests"]:
            if dest_arch_dict["platform"]["architecture"] not in src_archs:
                missing_archs.append(deepcopy(dest_arch_dict))
                missing_archs_log.append(dest_arch_dict["platform"]["architecture"])

        LOG.info(
            "Architectures missing from the new image: {0}".format(", ".join(missing_archs_log))
        )
        return missing_archs

    def _add_missing_architectures(self, src_manifest_list, missing_archs):
        """
        Add missing architectures to the source manifest list.

        Args:
            src_manifest_list (dict):
                Source manifest list.
            missing_archs ([dict]):
                Manifest data of missing architectures.
        Retuns (dict):
            New manifest list containing all the architectures.
        """
        new_manifest_list = deepcopy(src_manifest_list)
        new_manifest_list["manifests"] = new_manifest_list["manifests"] + missing_archs

        return new_manifest_list

    def merge_manifest_lists_selected_architectures(self, eligible_archs):
        """
        Merge manifests lists. Only specified archs are eligible for merging.

        This is an alternate workflow used in 'tag-docker'.

        Args:
            eligible_archs ([str]):
                Archs eligible for merging with the old manifest list.
        Returns (dict):
            New manifest list.
        """
        if not self._src_quay_client or not self._dest_quay_client:
            raise RuntimeError("QuayClient instance must be set for both source and dest images")

        src_manifest_list = self._src_quay_client.get_manifest(self.src_image, manifest_list=True)
        # It's possible that destination doesn't exist in this workflow. ML merging logic is still
        # necessary due to only some archs being eligible
        try:
            dest_manifest_list = self._dest_quay_client.get_manifest(
                self.dest_image, manifest_list=True
            )
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404 or e.response.status_code == 401:
                dest_manifest_list = None
            else:
                raise

        archs_to_add = []
        manifests_to_add = []
        for src_arch_dict in src_manifest_list["manifests"]:
            if src_arch_dict["platform"]["architecture"] in eligible_archs:
                manifests_to_add.append(deepcopy(src_arch_dict))
                archs_to_add.append(src_arch_dict["platform"]["architecture"])

        manifests_to_keep = []
        if dest_manifest_list:
            for dest_arch_dict in dest_manifest_list["manifests"]:
                if dest_arch_dict["platform"]["architecture"] not in archs_to_add:
                    manifests_to_keep.append(deepcopy(dest_arch_dict))

        new_manifest_list = deepcopy(src_manifest_list)
        new_manifest_list["manifests"] = manifests_to_add + manifests_to_keep

        return new_manifest_list



import base64
from datetime import datetime
import json
import logging
import uuid
import tempfile

import proton

from .exceptions import SigningError
from .utils.misc import run_entrypoint, log_step, get_pyxis_ssl_paths
from .quay_client import QuayClient
from .manifest_claims_handler import ManifestClaimsHandler

LOG = logging.getLogger("pubtools.quay")


class SignatureHandler:
    """Base class implementing operations common for container and operator signing."""

    MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST = 50

    def __init__(self, hub, task_id, target_settings, target_name):
        """
        Initialize.

        Args:
            hub (HubProxy):
                Instance of XMLRPC pub-hub proxy.
            task_id (str):
                ID of the pub task
            target_settings (dict):
                Target settings.
            target_name (str):
                Name of the target.
        """
        self.hub = hub
        self.task_id = task_id
        self.target_settings = target_settings
        self.target_name = target_name

        # Which URL hostnames will the destination images be accessible by to customers
        self.dest_registries = target_settings["docker_settings"]["docker_reference_registry"]
        self.dest_registries = (
            self.dest_registries
            if isinstance(self.dest_registries, list)
            else [self.dest_registries]
        )

        self.quay_host = self.target_settings.get("quay_host", "quay.io").rstrip("/")
        self._src_quay_client = None

    @property
    def src_quay_client(self):
        """Create and access QuayClient for source image."""
        if self._src_quay_client is None:
            self._src_quay_client = QuayClient(
                self.target_settings["source_quay_user"],
                self.target_settings["source_quay_password"],
                self.quay_host,
            )
        return self._src_quay_client

    @classmethod
    def create_manifest_claim_message(
        cls, destination_repo, signature_key, manifest_digest, docker_reference, image_name, task_id
    ):
        """
        Construct a manifest claim (image signature) as well as a message to send to RADAS.

        Constructed signature adheres to the following standard:
        https://github.com/containers/image/blob/master/docs/containers-signature.5.md

        Args:
            destination_repo (str):
                Internal destination repository to send to RADAS.
            signature_key (str):
                Signature key that will be sent to RADAS.
            manifest_digest (str):
                Digest referencing the signed image. Mandatory part of the image signature.
            docker_reference (str):
                Image reference which will be used by customers to pull the image. Mandatory part of
                the image signature.
            image_name (str):
                Name of the image to send to RADAS.
            task_id (str):
                ID of the pub task.
        """
        # container image signature
        manifest_claim = {
            "critical": {
                "type": "atomic container signature",
                "image": {"docker-manifest-digest": manifest_digest},
                "identity": {"docker-reference": docker_reference},
            },
            # NOTE: pub version is no longer written here. I hope that's OK
            "optional": {"creator": "Red Hat RCM Pub"},
        }

        message = {
            "sig_key_id": signature_key,
            # Python 2.6/3 compatibility workaround
            "claim_file": base64.b64encode(json.dumps(manifest_claim).encode("latin1")).decode(
                "latin1"
            ),
            "pub_task_id": task_id,
            "request_id": str(uuid.uuid4()),
            "manifest_digest": manifest_digest,
            "repo": destination_repo,
            "image_name": image_name,
            "docker_reference": docker_reference,
            "created": datetime.utcnow().isoformat() + "Z",
        }
        return message

    def get_tagged_image_digests(self, image_ref):
        """
        Get all digests referenced by a tagged image.

        There will only be one digest in case of single-arch image (source image), or multiple
        digests for a multi-arch image.

        Args:
            image_ref (str):
                Image reference URL. Must be specified via tag.

        Returns ([str]):
            List of manifest digests referenced by the tag.
        """
        digests = []

        manifest = self.src_quay_client.get_manifest(image_ref)
        # If V2S2 manifest, we only want its digest
        if manifest["mediaType"] == "application/vnd.docker.distribution.manifest.v2+json":
            digests.append(self.src_quay_client.get_manifest_digest(image_ref))
        # If manifest list, we want digests of all its arch images
        else:
            for arch_manifest in manifest["manifests"]:
                digests.append(arch_manifest["digest"])

        return digests

    def get_signatures_from_pyxis(self, manifest_digests=None):
        """
        Get existing signatures from Pyxis based on the specified criteria (currently only digests).

        NOTE: In the current implementation, only manifest digests are being used to search for
        existing signatures. Also, the search is performed in chunks, their size being limited by
        MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST.

        Args:
            manifest_digests ([str]|None):
                Digests for which to return signatures.

            Yields (dict):
                Existing signatures as returned by Pyxis based on specified criteria. The returned
                sturcture is an iterator to reduce memory requirements.
        """
        cert, key = get_pyxis_ssl_paths(self.target_settings)
        chunk_size = self.MAX_MANIFEST_DIGESTS_PER_SEARCH_REQUEST
        manifest_digests = sorted(list(set(manifest_digests)))

        for chunk_start in range(0, len(manifest_digests), chunk_size):
            chunk = manifest_digests[chunk_start : chunk_start + chunk_size]  # noqa: E203

            args = ["--pyxis-server", self.target_settings["pyxis_server"]]
            args += ["--pyxis-ssl-crtfile", cert]
            args += ["--pyxis-ssl-keyfile", key]

            with tempfile.NamedTemporaryFile(
                mode="w", prefix="pubtools_quay_get_signatures_"
            ) as signature_fetch_file:
                if manifest_digests:
                    json.dump(chunk, signature_fetch_file)
                    signature_fetch_file.flush()
                    args += ["--manifest-digest", "@{0}".format(signature_fetch_file.name)]

                env_vars = {}
                chunk_results = run_entrypoint(
                    ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-get-signatures"),
                    "pubtools-pyxis-get-signatures",
                    args,
                    env_vars,
                )

            for result in chunk_results:
                yield result

    def remove_duplicate_claim_messages(self, claim_messages):
        """
        Remove claim messages which could be considered 'duplicates', containing the same data.

        Args:
            claim_messages ([dict]):
                All created claim messages.

        Returns ([dict]):
            De-duplicated claim messages.
        """
        # dictionary key is a tuple of all parameters whose combination makes the message unique
        unique_message_mapping = {}
        for message in claim_messages:
            key = (
                message["sig_key_id"],
                message["claim_file"],
                message["pub_task_id"],
                message["manifest_digest"],
                message["repo"],
                message["image_name"],
                message["docker_reference"],
            )
            if key not in unique_message_mapping:
                unique_message_mapping[key] = message

        return list(unique_message_mapping.values())

    def filter_claim_messages(self, claim_messages):
        """
        Filter out the manifest claim messages which are already in the sigstore.

        Args:
            claim_messages ([dict]):
                Messages to be sent to RADAS.

        Returns ([dict]):
            Messages which don't yet exist in Pyxis.
        """
        LOG.info("Removing claim messages which already exist in Pyxis")
        digests = [message["manifest_digest"] for message in claim_messages]
        digests = sorted(list(set(digests)))

        existing_signatures = self.get_signatures_from_pyxis(manifest_digests=digests)

        signatures_by_key = {}
        for signature in existing_signatures:
            # combination of image reference, digest, and signature key makes a signature unique
            key = (signature["reference"], signature["manifest_digest"], signature["sig_key_id"])
            signatures_by_key[key] = signature

        filtered_claim_messages = []
        for message in claim_messages:
            key = (message["docker_reference"], message["manifest_digest"], message["sig_key_id"])
            # New signatures have switched to using long (16B) keys, while old signatures may still
            # contain short (8B) keys. If claim is matched with a shorter key format, it's
            # still considered a duplicate and shouldn't be uploaded again.
            old_key = None
            if len(message["sig_key_id"]) > 8:
                old_key = (
                    message["docker_reference"],
                    message["manifest_digest"],
                    message["sig_key_id"][-8:],
                )

            if key not in signatures_by_key and old_key not in signatures_by_key:
                filtered_claim_messages.append(message)

        LOG.info(
            "{0} claim messages remain after removing duplicates".format(
                len(filtered_claim_messages)
            )
        )
        return filtered_claim_messages

    def get_signatures_from_radas(self, claim_messages):
        """
        Send signature claims to RADAS via UMB and receive signed claims.

        The messaging logic is handled by the ManifestClaimsHandler class.

        Args:
            claim_messages ([dict]):
                Signature claims to be sent to RADAS.
        Returns ([dict]):
            Response messages from RADAS.
        raises MessageHandlerTimeoutException:
            If a message from RADAS hasn't arrived in time.
        """
        LOG.info("Sending claim messages to RADAS and waiting for results")
        # messages will be sent by pub-hub via XMLRPC
        # callback will be utilized by ManifestClaimsHandler, which will decide when to send msgs
        message_sender_callback = (
            lambda messages: self.hub.worker.umb_send_manifest_claim_messages(  # noqa: E731
                self.target_name, self.task_id, messages
            )
        )

        address = (
            "queue://Consumer.msg-producer-pub"
            ".{task_id}.VirtualTopic.eng.robosignatory.container.sign".format(task_id=self.task_id)
        )

        docker_settings = self.target_settings["docker_settings"]
        claims_handler = ManifestClaimsHandler(
            umb_urls=docker_settings["umb_urls"],
            radas_address=docker_settings.get("umb_radas_address", address),
            claim_messages=claim_messages,
            pub_cert=docker_settings.get("umb_pub_cert", "/etc/pub/umb-pub-cert-key.pem"),
            ca_cert=docker_settings.get("umb_ca_cert", "/etc/pki/tls/certs/ca-bundle.crt"),
            timeout=docker_settings.get("umb_signing_timeout", 600),
            throttle=docker_settings.get("umb_signing_throttle", 100),
            retry=docker_settings.get("umb_signing_retry", 3),
            message_sender_callback=message_sender_callback,
        )
        container = proton.reactor.Container(claims_handler)
        container.run()

        return claims_handler.received_messages

    def upload_signatures_to_pyxis(self, claim_mesages, signature_messages):
        """
        Upload signatures to Pyxis by using a pubtools-pyxis entrypoint.

        Data required for a Pyxis POST request:
        - manifest_digest
        - reference
        - repository
        - sig_key_id
        - signature_data

        Signatures are uploaded in batches.

        Args:
            claim_messages ([dict]):
                Signature claim messages constructed for the RADAS service.
            signature_messages ([dict]):
                Messages from RADAS containing image signatures.
        """
        LOG.info("Sending new signatures to Pyxis")

        cert, key = get_pyxis_ssl_paths(self.target_settings)

        signatures = []
        claim_messages_by_id = dict((m["request_id"], m) for m in claim_mesages)
        sorted_signature_messages = sorted(signature_messages, key=lambda msg: msg["request_id"])

        for signature_message in sorted_signature_messages:
            claim_message = claim_messages_by_id[signature_message["request_id"]]

            signatures.append(
                {
                    "manifest_digest": signature_message["manifest_digest"],
                    "reference": claim_message["docker_reference"],
                    "repository": claim_message["image_name"],
                    "sig_key_id": claim_message["sig_key_id"],
                    "signature_data": signature_message["signed_claim"],
                }
            )

        args = ["--pyxis-server", self.target_settings["pyxis_server"]]
        args += ["--pyxis-ssl-crtfile", cert]
        args += ["--pyxis-ssl-keyfile", key]

        with tempfile.NamedTemporaryFile(
            mode="w", prefix="pubtools_quay_upload_signatures_"
        ) as signature_file:
            json.dump(signatures, signature_file)
            signature_file.flush()
            args += ["--signatures", "@{0}".format(signature_file.name)]

            LOG.info("Uploading {0} new signatures".format(len(signatures)))
            env_vars = {}
            run_entrypoint(
                ("pubtools-pyxis", "console_scripts", "pubtools-pyxis-upload-signatures"),
                "pubtools-pyxis-upload-signature",
                args,
                env_vars,
            )

    def validate_radas_messages(self, claim_messages, signature_messages):
        """
        Check if messages received from RADAS contain any errors.

        Args:
            claim_messages ([dict]):
                Messages sent to RADAS.
            signature_messages ([dict]):
                Messages received from RADAS.

        Raises:
            SigningError:
                If RADAS messages contain errors.
        """
        failed_messages = 0
        for message in signature_messages:
            if message["errors"]:
                failed = [m for m in claim_messages if m["request_id"] == message["request_id"]][0]
                LOG.error(
                    "Signing of claim message {0} failed with following errors: {1}".format(
                        failed, message["errors"]
                    )
                )
                failed_messages += 1

        if failed_messages:
            raise SigningError(
                "Signing of {0}/{1} messages has failed".format(
                    failed_messages, len(claim_messages)
                )
            )


class ContainerSignatureHandler(SignatureHandler):
    """Class for handling the signing of container images."""

    def construct_item_claim_messages(self, push_item):
        """
        Construct all the signature claim messages for RADAS for one push item.

        push_item (ContainerPushItem):
            Container push item whose claim messages will be created.
        Returns ([dict]):
            Claim messages for a given push item.
        """
        LOG.info("Constructing claim messages for push item '{0}'".format(push_item))
        claim_messages = []

        if push_item.claims_signing_key:
            digests = self.get_tagged_image_digests(push_item.metadata["pull_url"])
            # each image digest needs its own signature
            for digest in digests:
                # each destination image reference needs its own signature
                for repo, tags in sorted(push_item.metadata["tags"].items()):
                    for tag in tags:
                        claim_messages += self.construct_variant_claim_messages(
                            repo, tag, digest, [push_item.claims_signing_key]
                        )

        return claim_messages

    def construct_variant_claim_messages(self, repo, tag, digest, signing_keys):
        """
        Construct claim messages for all specified variations of a given image.

        The variations are customer visible destination registry and signing key.

        Args:
            repo (str):
                Destination external repository  of a pushed image.
            tag: (str):
                Destination tag of a pushed image
            digest (str):
                Digest of the pushed image.
            signing_keys ([str]):
                Signing keys to construct the signatures with.

        Returns ([dict]):
            Signature claim messages to send to RADAS.
        """
        claim_messages = []
        image_schema = "{host}/{repository}:{tag}"

        for registry in self.dest_registries:
            reference = image_schema.format(host=registry, repository=repo, tag=tag)

            for signing_key in signing_keys:
                claim_message = self.create_manifest_claim_message(
                    destination_repo=repo,
                    signature_key=signing_key,
                    manifest_digest=digest,
                    docker_reference=reference,
                    image_name=repo,
                    task_id=self.task_id,
                )
                claim_messages.append(claim_message)

        return claim_messages

    @log_step("Sign container images")
    def sign_container_images(self, push_items):
        """
        Perform all the steps needed to sign the images of specified push items.

        The workflow can be summarized as:
        - create manifest claim messages for all items, registries, keys, digests, repos, and tags
        - filter out requests for signatures which are already in Pyxis
        - send messages to RADAS and receive signatures (ManifestClaimsHandler class)
        - Upload new signatures to Pyxis

        Args:
            push_items (([ContainerPushItem])):
                Container push items whose images will be signed.
        """
        if not self.target_settings["docker_settings"].get(
            "docker_container_signing_enabled", False
        ):
            LOG.info("Container signing not allowed in target settings, skipping.")
            return

        claim_messages = []
        for item in push_items:
            claim_messages += self.construct_item_claim_messages(item)
        claim_messages = self.remove_duplicate_claim_messages(claim_messages)
        claim_messages = self.filter_claim_messages(claim_messages)
        if not claim_messages:
            LOG.info("No new claim messages will be uploaded")
            return

        LOG.info("{0} claim messages will be uploaded".format(len(claim_messages)))
        signature_messages = self.get_signatures_from_radas(claim_messages)
        self.validate_radas_messages(claim_messages, signature_messages)
        self.upload_signatures_to_pyxis(claim_messages, signature_messages)


class OperatorSignatureHandler(SignatureHandler):
    """Class for handling the signing of index images."""

    def construct_index_image_claim_messages(self, index_image, tag, signing_keys):
        """
        Construct signature claim messages for RADAS for the specified index image.

        index_image (str):
            Reference to a new index image constructed by IIB.
        tag (str):
            Tag of the newly built index image.
        signing_keys (str):
            Signing keys to be used for signing.

        Returns ([dict]):
            Structured messages to be sent to UMB.
        """
        LOG.info("Constructing claim messages for index image '{0}'".format(index_image))
        claim_messages = []
        image_schema = "{host}/{repository}:{tag}"

        # Get digests of all archs this index image was build for
        index_image_credential = self.target_settings["iib_overwrite_from_index_token"].split(":")
        index_image_quay_client = QuayClient(
            index_image_credential[0],
            index_image_credential[1],
            self.quay_host,
        )
        manifest_list = index_image_quay_client.get_manifest(index_image, manifest_list=True)
        digests = [m["digest"] for m in manifest_list["manifests"]]
        for registry in self.dest_registries:
            for signing_key in signing_keys:
                if not signing_key:
                    continue
                for digest in digests:
                    repo = self.target_settings["quay_operator_repository"]
                    reference = image_schema.format(host=registry, repository=repo, tag=tag)
                    claim_message = self.create_manifest_claim_message(
                        destination_repo=repo,
                        signature_key=signing_key,
                        manifest_digest=digest,
                        docker_reference=reference,
                        image_name=repo,
                        task_id=self.task_id,
                    )
                    claim_messages.append(claim_message)

        return claim_messages

    @log_step("Sign operator images")
    def sign_operator_images(self, iib_results):
        """
        Perform all the steps needed to sign the newly constructed index images.

        Sigstore is not checked for existing signatures, as there's no way any could exist for a
        newly constructed image.

        Args:
            iib_results ({str:dict}):
                IIB results for each version the push was performed for.
        """
        image_schema = "{host}/{namespace}/{repo}@{digest}"
        if not self.target_settings["docker_settings"].get(
            "docker_container_signing_enabled", False
        ):
            LOG.info("Container signing not allowed in target settings, skipping.")
            return

        claim_messages = []
        for version, iib_details in sorted(iib_results.items()):
            iib_result = iib_details["iib_result"]
            signing_keys = iib_details["signing_keys"]
            # Using intermediate index image to ensure that it doesn't get overwritten
            iib_namespace = iib_result.index_image_resolved.split("/")[1]
            image_digest = iib_result.index_image_resolved.split("@")[1]
            intermediate_index_image = image_schema.format(
                host=self.target_settings.get("quay_host", "quay.io").rstrip("/"),
                namespace=iib_namespace,
                repo="iib",
                digest=image_digest,
            )
            # Version acts as a tag of the index image
            claim_messages += self.construct_index_image_claim_messages(
                intermediate_index_image, version, signing_keys
            )

        if not claim_messages:
            LOG.info("No new claim messages will be uploaded")
            return

        signature_messages = self.get_signatures_from_radas(claim_messages)
        self.validate_radas_messages(claim_messages, signature_messages)

        self.upload_signatures_to_pyxis(claim_messages, signature_messages)
        return claim_messages

    def sign_task_index_image(self, signing_keys, index_image, tag):
        """
        Perform an alternatve signing workflow used by IIB methods in pub.

        This workflow is used by methods 'PushAddIIBBundles', 'PushRemoveIIBOperators',
        'PushIIBBuildFromScratch'.

        Args:
            signing_keys ([str]):
                Signing key to be used.
            index_image (str):
                Index image pointing to the new manifest list.
            tag (str):
                Tag of the result index image.
        Returns ([dict]):
            Constructed claim messages.
        """
        claim_messages = self.construct_index_image_claim_messages(index_image, tag, signing_keys)
        if not claim_messages:
            LOG.info("No new claim messages will be uploaded")
            return

        signature_messages = self.get_signatures_from_radas(claim_messages)
        self.validate_radas_messages(claim_messages, signature_messages)

        self.upload_signatures_to_pyxis(claim_messages, signature_messages)

        return claim_messages


class BasicSignatureHandler(SignatureHandler):
    """Class that handles signing claims which were constructed by user."""

    def __init__(self, hub, target_settings, target_name):
        """
        Initialize.

        NOTE: "task_id" is not needed for this workflow

        Args:
            hub (HubProxy):
                Instance of XMLRPC pub-hub proxy.
            target_settings (dict):
                Target settings.
            target_name (str):
                Name of the target.
        """
        SignatureHandler.__init__(self, hub, "1", target_settings, target_name)

    def sign_claim_messages(self, claim_messages, remove_duplicates=True, filter_existing=True):
        """
        Sign claim messages that were provided by the user and upload them to Pyxis.

        Args:
            claim_messages ([dict]):
                Claim messages to be signed and uploaded.
            remove_duplicates (bool):
                Whether to check if there are any duplicates among the messages and remove them.
            filter_existing (bool):
                Whether to check if the signatures already exist in Pyxis, and only upload
                those that aren't.
        """
        if not self.target_settings["docker_settings"].get(
            "docker_container_signing_enabled", False
        ):
            LOG.info("Container signing not allowed in target settings, skipping.")
            return

        if remove_duplicates:
            claim_messages = self.remove_duplicate_claim_messages(claim_messages)
        if filter_existing:
            claim_messages = self.filter_claim_messages(claim_messages)
        if not claim_messages:
            LOG.info("No new claim messages will be uploaded")
            return

        LOG.info("{0} claim messages will be uploaded".format(len(claim_messages)))
        signature_messages = self.get_signatures_from_radas(claim_messages)
        self.validate_radas_messages(claim_messages, signature_messages)
        self.upload_signatures_to_pyxis(claim_messages, signature_messages)



from io import BytesIO
import json

from .quay_steps import (
    StepSanitizeContainerPushItems,
    StepPushContainerImgs,
    StepMergeManifestList,
    StepSignContainers,
    StepSanitizeRepositories,
    StepPushOperators,
    StepBuildBackupMapping,
    StepSanitizeOperatorPushItems,
    StepRollback,
)

from .utils.stepper import Stepper
from .utils.logger import Logger

import pushcollector

CLI_DESCRIPTION = ""


def log_push_items(signing_key, items=None):
    """Update push items according to their state."""
    collector = pushcollector.Collector.get()
    push_items_to_log = []
    for item in items:
        for repo in item.repos or [""]:
            push_items_to_log.append(
                {
                    "filename": item.file_name,
                    "src": item.file_path,
                    "dest": repo,
                    "build": item.build,
                    "state": item.state,
                    "origin": item.origin,
                    "signing_key": signing_key.upper(),
                    "checksums": item.checksums,
                }
            )
    collector.update_push_items(push_items_to_log)


def push_docker(push_items, signing_key, hub, task_id, target_name, target_settings):
    """Run push docker sequence.

    Arguments:
        push_items (<list-of-push-items>)
            List of push items containing metadata
        signing_key (str)
            Signing key which should be used to sign pushed manifests
        hub (xmlrpc.client.ServerProxy)
            Pubhub XMLRPC proxy instance
        task_id (int)
            Pub task id
        target_name (str)
            Pub target name
        target_settings (dict)
            Target settings
    """
    log_push_items(signing_key, items=push_items)
    shared_data = {}
    logger = Logger()
    common_external_res = {
        "push_items": push_items,
        "log_info": logger.log_info,
        "log_error": logger.log_error,
        "log_warning": logger.log_warning,
        "log_debug": logger.log_debug,
        "hub": hub,
        "task_id": task_id,
        "target_name": target_name,
    }

    stepper = Stepper(shared_data)
    stepper.add_step(
        StepSanitizeContainerPushItems(
            "1", (), {}, shared_data, external_resources=common_external_res
        )
    )
    stepper.add_step(
        StepSanitizeRepositories(
            "1",
            ("StepSanitizeContainerPushItems:1",),
            {},
            shared_data,
            external_resources=common_external_res,
        )
    )
    stepper.add_step(
        StepBuildBackupMapping(
            "1",
            ("StepSanitizeContainerPushItems:1",),
            {},
            shared_data,
            external_resources=common_external_res,
        )
    )
    stepper.add_step(
        StepPushContainerImgs(
            "1",
            ("StepSanitizeContainerPushItems:1", target_settings),
            {},
            shared_data,
            external_resources=common_external_res,
        )
    )
    stepper.add_step(
        StepMergeManifestList(
            "1",
            ("StepSanitizeContainerPushItems:1", target_settings),
            {},
            shared_data,
            external_resources=common_external_res,
        )
    )
    stepper.add_step(
        StepSignContainers(
            "1",
            ("StepSanitizeContainerPushItems:1", target_settings),
            {
                "container_signing_enabled": target_settings["docker_settings"].get(
                    "container_signing_enabled", False
                )
            },
            shared_data,
            external_resources=common_external_res,
        )
    )
    stepper.add_step(
        StepSanitizeOperatorPushItems(
            "1",
            (),
            {"auto_upload_operators": target_settings.get("auto_upload_operators")},
            shared_data,
            external_resources=common_external_res,
        )
    )
    stepper.add_step(
        StepPushOperators(
            "1",
            ("StepSanitizeOperatorPushItems:1", target_settings),
            {"docker_reference_registry": target_settings.get("docker_reference_registry")},
            shared_data,
            external_resources=common_external_res,
        )
    )
    stepper.add_step(
        StepSignContainers(
            "index-image",
            ("StepSanitizeContainerPushItems:1",),
            {
                "autoupload_operators": target_settings.get("auto_upload_operators"),
                "docker_reference_registry": target_settings.get("docker_reference_registry"),
                "iib_server": target_settings["iib_server"],
            },
            shared_data,
            external_resources=common_external_res,
        )
    )
    try:
        stepper.run()
    except Exception:
        stepper.add_step(
            StepRollback(
                "1",
                ("StepBuildBackupMapping:1", target_settings),
                {},
                shared_data,
                external_resources=common_external_res,
            )
        )
        stepper.run(start_from=-1)
        raise
    finally:
        log_push_items(signing_key, items=push_items)
        results = stepper.dump()
        json_io = BytesIO(str(json.dumps(results) + "\n").encode("utf-8"))
        hub.upload_task_log(json_io, task_id, "report.json")
    return stepper.shared_results


def mod_entry_point(push_items, hub, task_id, target_name, target_settings):
    """Entry point for use in another python code."""
    return push_docker(push_items, "signing-key", hub, task_id, target_name, target_settings)



import logging

from pubtools.pluggy import pm, task_context

from .image_untagger import ImageUntagger
from .utils.misc import setup_arg_parser, add_args_env_variables, send_umb_message

LOG = logging.getLogger("pubtools.quay")

UNTAG_IMAGES_ARGS = {
    ("--reference",): {
        "help": "Image reference to untag. Must be specified by tag. Multiple can be specified.",
        "required": True,
        "type": str,
        "action": "append",
    },
    ("--remove-last",): {
        "help": "Whether to remove a tag even if it's the last reference of some image.",
        "required": False,
        "type": bool,
    },
    ("--quay-api-token",): {
        "help": "OAuth token for Quay REST API.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_API_TOKEN",
    },
    ("--quay-user",): {
        "help": "Username for Quay login.",
        "required": False,
        "type": str,
    },
    ("--quay-password",): {
        "help": "Password for Quay. Can be specified by env variable QUAY_PASSWORD.",
        "required": False,
        "type": str,
        "env_variable": "QUAY_PASSWORD",
    },
    ("--send-umb-msg",): {
        "help": "Flag of whether to send a UMB message",
        "required": False,
        "type": bool,
    },
    ("--umb-url",): {
        "help": "UMB URL. More than one can be specified.",
        "required": False,
        "type": str,
        "action": "append",
    },
    ("--umb-cert",): {
        "help": "Path to the UMB certificate for SSL authentication.",
        "required": False,
        "type": str,
    },
    ("--umb-client-key",): {
        "help": "Path to the UMB private key for accessing the certificate.",
        "required": False,
        "type": str,
    },
    ("--umb-ca-cert",): {
        "help": "Path to the UMB CA certificate.",
        "required": False,
        "type": str,
    },
    ("--umb-topic",): {
        "help": "UMB topic to send the message to.",
        "required": False,
        "type": str,
        "default": "VirtualTopic.eng.pub.quay_untag_image",
    },
}


def construct_kwargs(args):
    """
    Construct a kwargs dictionary based on the entered command line arguments.

    Args:
        args (argparse.Namespace):
            Parsed command line arguments.

    Returns (dict):
        Keyword arguments for the 'untag_images' function.
    """
    kwargs = args.__dict__

    # in args.__dict__ unspecified bool values have 'None' instead of 'False'
    for name, attributes in UNTAG_IMAGES_ARGS.items():
        if attributes["type"] is bool:
            bool_var = name[0].lstrip("-").replace("-", "_")
            if kwargs[bool_var] is None:
                kwargs[bool_var] = False

    # some exceptions have to be remapped
    kwargs["references"] = kwargs.pop("reference")
    kwargs["umb_urls"] = kwargs.pop("umb_url")

    return kwargs


def verify_untag_images_args(
    references, quay_user, quay_password, send_umb_msg, umb_urls, umb_cert
):
    """
    Verify the presence and correctness of input parameters.

    Args:
        references ([str]):
            List of image references to untag.
        quay_user (str):
            Quay username for Docker HTTP API.
        quay_password (str):
            Quay password for Docker HTTP API.
        send_umb_msg (bool):
            Whether to send UMB messages about the untagged images.
        umb_urls ([str]):
            AMQP broker URLs to connect to.
        umb_cert (str):
            Path to a certificate used for UMB authentication.
    """
    for reference in references:
        if "@" in reference:
            raise ValueError("All references must be specified via tag, not digest")

    if (quay_user and not quay_password) or (quay_password and not quay_user):
        raise ValueError("Both user and password must be present when attempting to log in.")

    if send_umb_msg:
        if not umb_urls:
            raise ValueError("UMB URL must be specified if sending a UMB message was requested.")
        if not umb_cert:
            raise ValueError(
                "A path to a client certificate must be provided when sending a UMB message."
            )


def untag_images(
    references,
    quay_api_token,
    remove_last=False,
    quay_user=None,
    quay_password=None,
    send_umb_msg=False,
    umb_urls=[],
    umb_cert=None,
    umb_client_key=None,
    umb_ca_cert=None,
    umb_topic="VirtualTopic.eng.pub.quay_untag_image",
):
    """
    Untag images from Quay.

    Args:
        references ([str]):
            List of image references to untag.
        quay_api_token (str):
            OAuth token for authentication of Quay REST API.
        remove_last (bool):
            Whether to remove a tag when it's the last reference of an image (in that repo).
        quay_user (str):
            Quay username for Docker HTTP API.
        quay_password (str):
            Quay password for Docker HTTP API.
        send_umb_msg (bool):
            Whether to send UMB messages about the untagged images.
        umb_urls ([str]):
            AMQP broker URLs to connect to.
        umb_cert (str):
            Path to a certificate used for UMB authentication.
        umb_client_key (str):
            Path to a client key to decrypt the certificate (if necessary).
        umb_ca_cert (str):
            Path to a CA certificate (for mutual authentication).
        umb_topic (str):
            Topic to send the UMB messages to.
    """
    verify_untag_images_args(references, quay_user, quay_password, send_umb_msg, umb_urls, umb_cert)

    LOG.info("Started untagging operation with the following references: {0}".format(references))
    untagger = ImageUntagger(references, quay_api_token, remove_last, quay_user, quay_password)
    lost_images = untagger.untag_images()

    LOG.info("Untagging operation succeeded")
    pm.hook.quay_images_untagged(untag_refs=sorted(references), lost_refs=sorted(lost_images))

    if send_umb_msg:
        LOG.info("Sending a UMB message")
        props = {"untag_refs": references, "lost_refs": lost_images}
        send_umb_message(
            umb_urls,
            props,
            umb_cert,
            umb_topic,
            client_key=umb_client_key,
            ca_cert=umb_ca_cert,
        )


def setup_args():
    """Set up argparser without extra parameters, this method is used for auto doc generation."""
    return setup_arg_parser(UNTAG_IMAGES_ARGS)


def untag_images_main(sysargs=None):
    """Entrypoint for untagging images."""
    logging.basicConfig(level=logging.INFO)

    parser = setup_args()
    if sysargs:
        args = parser.parse_args(sysargs[1:])
    else:
        args = parser.parse_args()  # pragma: no cover"
    args = add_args_env_variables(args, UNTAG_IMAGES_ARGS)

    if not args.quay_api_token:
        raise ValueError("--quay-api-token must be specified")

    kwargs = construct_kwargs(args)

    with task_context():
        untag_images(**kwargs)
